from __future__ import annotations

from collections.abc import Callable
from typing import List, Generator, TYPE_CHECKING, Literal, Optional, Union

from simpy import events
import simpy
import logging

from prodsys.models.port_data import StoreData
from prodsys.models.resource_data import ResourceData
from prodsys.models import port_data
from prodsys.simulation.process_handlers.disassembly_process_handler import DisassemblyProcessHandler
from prodsys.simulation.request import Request
from prodsys.simulation.entities.entity import Entity
from prodsys.models.dependency_data import DependencyType
from prodsys.simulation import (
    sim,
    process,
)
from prodsys.simulation.process_handlers.production_process_handler import ProductionProcessHandler
from prodsys.simulation.process_handlers.transport_process_handler import TransportProcessHandler, ConveyorTransportProcessHandler
from prodsys.simulation.process_handlers.dependency_process_handler import DependencyProcessHandler
from prodsys.simulation.process_handlers.system_process_model_process_handler import SystemProcessModelHandler
from prodsys.simulation.process_handlers.resource_process_model_process_handler import ResourceProcessModelHandler
from prodsys.models.resource_data import ResourceType
from prodsys.simulation import request as request_module

if TYPE_CHECKING:
    from prodsys.simulation import (
        process,
        resources,
    )
    from prodsys.simulation import request as request_module
    from prodsys.control import sequencing_control_env
    from prodsys.simulation.locatable import Locatable
    from prodsys.simulation.lot_handler import LotHandler

logger = logging.getLogger(__name__)


class Controller:
    """
    A controller is responsible for controlling the processes of a resource. The controller is requested by products requiring processes. The controller decides has a control policy that determines with which sequence requests are processed.

    Args:
        control_policy (Callable[[List[Request]], None]): The control policy that determines the sequence of requests to be processed.
        env (sim.Environment): The environment in which the controller is running.

    Attributes:
        control_policy (Callable[[List[Request]], None]): The control policy that determines the sequence of requests to be processed.
        env (sim.Environment): The environment in which the controller is running.
        requests (List[Request]): The list of requests that are waiting to be processed.
        requested (events.Event): The event that is triggered when a request is made.
        resource (resources.Resource): The resource that is controlled by the controller.
        process_finished (events.Event): The event that is triggered when a process is finished.
        num_running_processes (int): The number of processes that are currently running.
        reserved_requests_count (int): The number of requests that are reserved for processing.
    """

    def __init__(
        self,
        control_policy: Callable[
            [
                List[request_module.Request],
            ],
            None,
        ],
        env: sim.Environment,
        lot_handler: LotHandler,
    ) -> None:
        self.control_policy = control_policy
        self.env = env
        self.lot_handler = lot_handler
        self.requests: List[request_module.Request] = []
        self.state_changed: events.Event = events.Event(env)
        self.resource: resources.Resource = None
        self.num_running_processes = 0
        self.reserved_requests_count = 0

    def set_resource(self, resource: resources.Resource) -> None:
        self.resource = resource
        self.env = resource.env

    def request(self, process_request: request_module.Request) -> None:
        """
        Request the controller consider the request in the future for processing.

        Args:
            process_request (Request): The request to be processed.
        """
        self.requests.append(process_request)
        if not self.state_changed.triggered:
            self.state_changed.succeed()

    def free_up_queue_check(self) -> Generator:
        # generator that runs until one output queue is free again, getting to know it from a get from the output queue
        output_queues = [port for port in self.resource.ports if port.data.interface_type == port_data.PortInterfaceType.OUTPUT]
        queue_get_events = [queue.on_space for queue in output_queues]
        yield simpy.AnyOf(self.env, queue_get_events)
        if not self.state_changed.triggered:
            self.state_changed.succeed()

    def control_loop(self) -> Generator:
        """
        The control loop is the main process of the controller. It has to run indefinetely.
        It should repeatedly check if requests are made or a process is finished and then start the next process.
        """
        while True:
            if self.resource.requires_charging:
                # TODO: transport AGV to charging station, -> use a ChargingHandler for this!
                yield self.env.process(self.resource.charge())
            yield self.state_changed
            self.state_changed = events.Event(self.env)
            if (
                self.resource.full
                or self.resource.in_setup
                or self.resource.bound
                or not self.requests
            ):
                continue

            def get_requests_with_available_dependencies(requests: List[request_module.Request]) -> List[request_module.Request]:
                requests_with_available_dependencies = []
                for request in requests:
                    if request.required_dependencies:
                        primitive_dependencies = [dependency for dependency in request.required_dependencies if dependency.data.dependency_type == DependencyType.ASSEMBLY]
                        if primitive_dependencies:
                            router = request.requesting_item.router
                            # Check if all required primitives are available
                            all_primitives_available = True
                            for dependency in primitive_dependencies:
                                required_primitive = dependency.required_entity
                                if required_primitive is None:
                                    all_primitives_available = False
                                    break
                                # Get the type from the required_primitive (can be Product or Primitive)
                                primitive_type = required_primitive.data.type
                                free_primitives = router.free_primitives_by_type.get(primitive_type, [])
                                # Check if there are actually free primitives (list is not empty)
                                if not free_primitives or len(free_primitives) == 0:
                                    all_primitives_available = False
                                    break
                            if not all_primitives_available:
                                continue
                    requests_with_available_dependencies.append(request)
                return requests_with_available_dependencies
            possible_requests = get_requests_with_available_dependencies(self.requests)
            if not possible_requests:
                continue
            self.control_policy(possible_requests)
            def is_request_feasible(request: request_module.Request) -> bool:
                # Check transport requests for target queue availability

                if request.request_type == request_module.RequestType.TRANSPORT:
                    if request.target_queue.is_full:
                        return False
                # Check production requests for INPUT_OUTPUT queue deadlock prevention
                elif request.request_type in (request_module.RequestType.PRODUCTION, request_module.RequestType.PROCESS_MODEL):
                    # For INPUT_OUTPUT queues, check if output space is available
                    # If origin == target (same INPUT_OUTPUT queue), check if item is in queue
                    if request.origin_queue == request.target_queue:
                        # If item is in queue, we can remove it then put it back
                        item_id = request.entity.data.ID
                        if item_id in request.origin_queue.items:
                            # Item is in queue - feasible (we'll remove then put back)
                            return True
                        else:
                            raise ValueError(f"Item {item_id} not in queue {request.origin_queue.data.ID}")
                    else:
                        # Separate queues - first check if item is in origin queue
                        item_id = request.entity.data.ID
                        is_in_origin = item_id in request.origin_queue.items
                        if not is_in_origin:
                            raise ValueError(f"Item {item_id} not in origin queue {request.origin_queue.data.ID}")
                        # Item is in origin queue - check if target has space
                        # if request.target_queue.is_full:
                        #     return False
                return True

            def get_feasible_request(requests: List[request_module.Request]) -> request_module.Request:
                for i, request in enumerate(requests):
                    if is_request_feasible(request):
                        return request # If request becomes infeasible (queue full), reroute it back to router
                    # This allows it to be retried later when space becomes available
                    # Only reroute transport requests - production requests should have been validated
                    # before routing and if item is in INPUT_OUTPUT queue, it should be processable
                    # if request.request_type == request_module.RequestType.TRANSPORT:
                    #     self.requests.remove(request)
                    #     request.requesting_item.router.request_handler.reroute_request(request)
                    #     # Trigger router to check for new routing opportunities
                    #     if not request.requesting_item.router.got_requested.triggered:
                    #         request.requesting_item.router.got_requested.succeed()
                return None
            
            selected_request = get_feasible_request(possible_requests)
            if not selected_request:
                # If there are requests waiting on full output queues, wait for space
                # self.env.process(self.free_up_queue_check())
                continue
            self.requests.remove(selected_request)
            if self._should_form_lot(selected_request):
                lot_request = self._form_lot(selected_request)
                if not lot_request:
                    # Can't form lot yet - move to end and try next request
                    self.requests.append(selected_request)
                    continue
                selected_request = lot_request
                
            # Reserve output queue for transport requests (production requests reserve in their handler)
            if selected_request.request_type == request_module.RequestType.TRANSPORT:
                self.reserve_output_queue(selected_request)
            
            self.reserve_resource_capacity(selected_request.capacity_required)
            # For dependency requests, immediately bind the resource to block other processes
            if selected_request.request_type in (request_module.RequestType.PROCESS_DEPENDENCY, request_module.RequestType.RESOURCE_DEPENDENCY):
                self.resource.bind_to_dependant(selected_request.requesting_item)
            self.resource.update_full()
            process_handler = get_requets_handler(selected_request)
            self.env.process(process_handler.handle_request(selected_request))
            if not self.resource.full and self.requests and not self.state_changed.triggered:
                self.state_changed.succeed()

    def reserve_output_queue(self, process_request: request_module.Request) -> Generator:
        """
        Reserve the output queue for the process.

        Args:
            process_request (request_module.Request): The request to reserve the output queue for.
        """
        for entity in process_request.get_atomic_entities():
            if process_request.target_queue.is_full:
                raise ValueError(f"Target queue {process_request.target_queue.data.ID} is full for request {process_request.completed}")
            process_request.target_queue.reserve()

    def _should_form_lot(self, process_request: request_module.Request) -> bool:
        return self.lot_handler.lot_required(process_request)

    def _form_lot(self, process_request: request_module.Request) -> Optional[request_module.Request]:
        if not self.lot_handler.is_lot_feasible(process_request):
            return None
        lot_requests = self.lot_handler.get_lot_request(process_request)
        return lot_requests

    def reserve_resource_capacity(self, capacity: int) -> None:
        """
        Reserve the resource capacity for the process.

        Args:
            process_request (request_module.Request): The request to reserve the resource capacity for.
        """
        if capacity > self.resource.get_free_capacity():
            raise ValueError(f"Resource {self.resource.data.ID} has not enough capacity to reserve {capacity}, current capacity: {self.resource.get_free_capacity()}, requested capacity: {capacity}")
        self.reserved_requests_count += capacity

    def unreserve_resource_capacity(self, capacity: int) -> None:
        """
        Unreserve the resource capacity for the process.

        Args:
            process_request (request_module.Request): The request to unreserve the resource capacity for.
        """
        self.reserved_requests_count -= capacity
        if self.reserved_requests_count < 0:
            raise ValueError(f"Resource {self.resource.data.ID} has not enough reserved to unreserve {capacity}, current capacity: {self.resource.get_free_capacity()}, requested capacity: {capacity}")

    def mark_started_process(self, num_processes: int = 1) -> None:
        """
        Mark the process as started.

        Args:
            num_processes (int): The number of processes that are being started.
        """
        self.unreserve_resource_capacity(num_processes)
        self.num_running_processes += num_processes

    def mark_finished_process(self, num_processes: int = 1) -> None:
        """
        Mark the process as finished.

        Args:
            process_request (Request): The request that is being processed.
        """
        self.num_running_processes -= num_processes
        self.resource.update_full()
        if not self.state_changed.triggered:
            self.state_changed.succeed()
            
    def mark_finished_process_no_sink_transport(self, process_request: Request, entity: Entity) -> None:
        """
        Mark the process as finished, but no transport to sink required.

        Args:
            process_request (Request): The request that is being processed.
        """
        
        self.num_running_processes -= process_request.capacity_required
        self.resource.update_full()
        entity.no_transport_to_sink = True
        if not self.state_changed.triggered:
            self.state_changed.succeed()


def get_requets_handler(
    request: request_module.Request,
) -> Union[ProductionProcessHandler, TransportProcessHandler, DependencyProcessHandler, SystemProcessModelHandler, ResourceProcessModelHandler]:
    """
    Get the process handler for a given process.

    Args:
        process (process.PROCESS_UNION): The process to get the handler for.

    Returns:
        Union[ProductionProcessHandler, TransportProcessHandler]: The process handler for the given process.
    """
    if (
        request.request_type == request_module.RequestType.PRODUCTION
        and any(dependency.data.dependency_type == DependencyType.DISASSEMBLY for dependency in request.required_dependencies)
    ):
        return DisassemblyProcessHandler(request.requesting_item.env) 
    elif (
        request.request_type == request_module.RequestType.PRODUCTION
        or request.request_type == request_module.RequestType.REWORK
    ):
        return ProductionProcessHandler(request.requesting_item.env)
    elif request.request_type == request_module.RequestType.TRANSPORT:
        if request.get_resource().can_move:
            return TransportProcessHandler(request.requesting_item.env)
        else:
            return ConveyorTransportProcessHandler(request.requesting_item.env)
    elif (
        request.request_type == request_module.RequestType.PROCESS_DEPENDENCY
        or request.request_type == request_module.RequestType.RESOURCE_DEPENDENCY
    ):
        return DependencyProcessHandler(request.requesting_item.env)
    elif request.request_type == request_module.RequestType.PROCESS_MODEL:
        # Route to SystemProcessModelHandler for system resources, ResourceProcessModelHandler for regular resources
        if request.resource.data.resource_type == ResourceType.SYSTEM:
            return SystemProcessModelHandler(request.requesting_item.env)
        else:
            return ResourceProcessModelHandler(request.requesting_item.env)
    else:
        raise ValueError(f"Unknown process type: {type(process)}")




def FIFO_control_policy(requests: List[request_module.Request]) -> None:
    """
    Sort the requests according to the FIFO principle.

    Args:
        requests (List[Request]): The list of requests.
    """
    pass


def LIFO_control_policy(requests: List[request_module.Request]) -> None:
    """
    Sort the requests according to the LIFO principle (reverse the list).

    Args:
        requests (List[Request]): The list of requests.
    """
    requests.reverse()


def SPT_control_policy(requests: List[request_module.Request]) -> None:
    """
    Sort the requests according to the SPT principle (shortest process time first).

    Args:
        requests (List[Request]): The list of requests.
    """
    requests.sort(key=lambda x: x.process.get_expected_process_time())


def get_location(locatable: Locatable) -> List[float]:
    return locatable.get_location()


def SPT_transport_control_policy(
    requests: List[request_module.Request],
) -> None:
    """
    Sort the requests according to the SPT principle (shortest process time first).

    Args:
        requests (List[request.Request]): The list of requests.
    """
    # for request in requests:
    #     if request.origin_queue is None or request.target_queue is None:
    #         raise ValueError(f"Origin queue or target queue is None for request {request.completed}")

    def get_expected_time(request: request_module.Request) -> float:
        if request.request_type in (request_module.RequestType.PROCESS_DEPENDENCY, request_module.RequestType.RESOURCE_DEPENDENCY):
            #  TODO: calculate time based on dependency process time
            return 0.1
        return request.process.get_expected_process_time(
            get_location(request.origin_queue), get_location(request.target_queue)
        )
    requests.sort(
        key=get_expected_time
    )


def nearest_origin_and_longest_target_queues_transport_control_policy(
    requests: List[request_module.Request],
) -> None:
    """
    Sort the requests according to nearest origin without considering the target location.
    Second order sorting by descending length of the target output queues, to prefer targets where a product can be picked up.
    Args:
        requests (List[request.Request]): The list of requests.
    """
    requests.sort(
        key=lambda x: (
            x.process.get_expected_process_time(
                get_location(x.resource), get_location(x.origin_queue)
            ),
            -x.target_queue.free_space(),
        )
    )


def nearest_origin_and_shortest_target_input_queues_transport_control_policy(
    requests: List[request_module.Request],
) -> None:
    """
    Sort the requests according to nearest origin without considering the target location.
    Second order sorting by ascending length of the target input queue so that resources with empty input queues get material to process.

    Args:
        requests (List[request.Request]): The list of requests.
    """
    requests.sort(
        key=lambda x: (
            x.process.get_expected_process_time(
                get_location(x.resource), get_location(x.origin_queue)
            ),
            x.target_queue.free_space(),
        )
    )


def agent_control_policy(
    gym_env: sequencing_control_env.AbstractSequencingControlEnv,
    requests: List[request_module.Request],
) -> None:
    """
    Sort the requests according to the agent's policy.

    Args:
        gym_env (gym_env.ProductionControlEnv): A gym environment, where the agent can interact with the simulation.
        requests (List[Request]): The list of requests.
    """
    gym_env.interrupt_simulation_event.succeed()

def scheduled_control_policy(
    schedule_sequence: List[tuple[str, str, int]], fallback_policy: Callable, requests: List[request_module.Request]
) -> None:
    """
    A control policy that sequences products based on their scheduled index in the sequence. 
    Matches requests based on both product ID and process ID to handle cases where the same
    product visits the resource multiple times with different processes.
    
    Matches each request to the next occurrence in the schedule sequence, ensuring that
    products are processed in the correct order even when they appear multiple times.
    
    If the request contains any product+process combination which is not in the schedule, 
    the fallback policy is also considered if it would be processed next. 

    Args:
        schedule_sequence (List[tuple[str, str, int]]): Ordered list of (product_id, process_id, index) 
            tuples representing the scheduled sequence
        fallback_policy (Callable): fallback control policy
        requests (List[request_module.Request]): list of requests to sequence
    """

    # Track which schedule entries have been matched in this call
    matched_schedule_indices = set()
    request_to_priority = {}
    non_scheduled_requests = []
    
    # Debug: Print incoming requests
    for i, request_instance in enumerate(requests):
        product_id = request_instance.entity.data.ID if request_instance.entity else "None"
        process_id = request_instance.process.data.ID if request_instance.process else "None"
    
    # Track which schedule entries have been matched and which products have been seen
    # We need to match requests optimally, considering all requests together
    # Strategy: For each product, track the next expected schedule index
    product_next_expected_index = {}
    
    # First pass: Build a map of all possible matches for each request
    request_matches = {}  # request -> list of (schedule_index, product_id, process_id) tuples
    for request_instance in requests:
        product_id = request_instance.entity.data.ID
        process_id = request_instance.process.data.ID if request_instance.process else None
        
        if not process_id:
            # print(f"[SCHEDULED_CONTROL] Request for {product_id} has no process ID - adding to non-scheduled")
            non_scheduled_requests.append(request_instance)
            continue
        
        # Find all possible matches for this request
        possible_matches = []
        for sched_product_id, sched_process_id, sched_index in schedule_sequence:
            if product_id == sched_product_id and process_id == sched_process_id:
                possible_matches.append((sched_index, sched_product_id, sched_process_id))
        
        if possible_matches:
            request_matches[request_instance] = possible_matches
            # print(f"[SCHEDULED_CONTROL] Request {product_id}->{process_id} has {len(possible_matches)} possible schedule matches: {[m[0] for m in possible_matches]}")
        else:
            non_scheduled_requests.append(request_instance)
    
    # Second pass: Match requests to schedule entries optimally
    # Strategy: Process schedule entries in order, and match requests as we go
    # This ensures that earlier schedule entries are matched first, maintaining correct order
    
    # Create a list of (request, possible_matches) sorted by earliest possible schedule index
    request_match_list = []
    for request_instance, possible_matches in request_matches.items():
        if possible_matches:
            earliest_index = min(m[0] for m in possible_matches)
            request_match_list.append((earliest_index, request_instance, possible_matches))
    
    # Sort by earliest possible schedule index
    request_match_list.sort(key=lambda x: x[0])
    
    # Now match requests in order of their earliest possible schedule index
    for earliest_index, request_instance, possible_matches in request_match_list:
        product_id = request_instance.entity.data.ID
        process_id = request_instance.process.data.ID
        
        # Get the next expected index for this product
        next_expected = product_next_expected_index.get(product_id, 0)
        
        # Find the earliest match that is >= next_expected and not already matched
        best_match = None
        for sched_index, sched_product_id, sched_process_id in possible_matches:
            if sched_index >= next_expected and sched_index not in matched_schedule_indices:
                if best_match is None or sched_index < best_match[0]:
                    best_match = (sched_index, sched_product_id, sched_process_id)
        
        if best_match:
            sched_index, _, _ = best_match
            request_to_priority[request_instance] = sched_index
            matched_schedule_indices.add(sched_index)
            # Update next expected index for this product
            product_next_expected_index[product_id] = sched_index + 1
        else:
            # No valid match found - this shouldn't happen if schedule is correct
            non_scheduled_requests.append(request_instance)
    
    # Debug: Print priorities assigned
    for request_instance, priority in sorted(request_to_priority.items(), key=lambda x: x[1]):
        product_id = request_instance.entity.data.ID
        process_id = request_instance.process.data.ID if request_instance.process else "None"
    
    # If all requests are scheduled, sort by priority
    if len(non_scheduled_requests) == 0:
        # All requests are in schedule - sort by their schedule priority
        requests.sort(key=lambda r: request_to_priority.get(r, float('inf')))
        for i, r in enumerate(requests):
            product_id = r.entity.data.ID
            process_id = r.process.data.ID if r.process else "None"
            priority = request_to_priority.get(r, "N/A")
        return
    
    # Some requests are not scheduled - check if fallback would help
    request_list_copy = requests[::]
    fallback_policy(request_list_copy)
    # If fallback would select a non-scheduled request, use fallback for all
    if request_list_copy[0] in non_scheduled_requests:
        fallback_policy(requests)
        return
    
    # Mix scheduled and non-scheduled: prioritize scheduled ones
    scheduled_requests = [r for r in requests if r in request_to_priority]
    scheduled_requests.sort(key=lambda r: request_to_priority[r])
    
    # Put scheduled requests first, then non-scheduled (sorted by fallback)
    requests.clear()
    requests.extend(scheduled_requests)
    requests.extend(non_scheduled_requests)
    
    for i, r in enumerate(requests):
        product_id = r.entity.data.ID
        process_id = r.process.data.ID if r.process else "None"
        priority = request_to_priority.get(r, "N/A (non-scheduled)")


# TODO: add a Controller which starts processes with delays...




"""
Tests for production_system_data module validation.
"""

import pytest
from pydantic import ValidationError
from prodsys.models.production_system_data import ProductionSystemData
from prodsys.models import (
    time_model_data,
    state_data,
    processes_data,
    port_data,
    resource_data,
    product_data,
    sink_data,
    source_data,
    node_data,
)


class TestFieldValidators:
    """Tests for field validators in ProductionSystemData."""

    def test_check_states_invalid_time_model(self):
        """Test that states with invalid time models raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        state = state_data.BreakDownStateData(
            ID="state1",
            description="State 1",
            time_model_id="invalid_tm",  # Invalid time model ID
            type="BreakDownState",
            repair_time_model_id="tm1",
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                state_data=[state],
            )
        
        assert "time model" in str(exc_info.value).lower()
        assert "state1" in str(exc_info.value)

    def test_check_processes_invalid_time_model(self):
        """Test that processes with invalid time models raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="invalid_tm",  # Invalid time model ID
            type="ProductionProcesses",
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process],
            )
        
        assert "time model" in str(exc_info.value).lower()
        assert "P1" in str(exc_info.value)

    def test_check_resources_invalid_process(self):
        """Test that resources with invalid processes raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["invalid_process"],  # Invalid process ID
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process],
                resource_data=[resource],
            )
        
        assert "process" in str(exc_info.value).lower()
        assert "R1" in str(exc_info.value)

    def test_check_resources_invalid_state(self):
        """Test that resources with invalid states raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        state = state_data.BreakDownStateData(
            ID="state1",
            description="State 1",
            time_model_id="tm1",
            type="BreakDownState",
            repair_time_model_id="tm1",
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            state_ids=["invalid_state"],  # Invalid state ID
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                state_data=[state],
                process_data=[process],
                resource_data=[resource],
            )
        
        assert "state" in str(exc_info.value).lower()
        assert "R1" in str(exc_info.value)

    def test_check_resources_invalid_port(self):
        """Test that resources with invalid ports raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        queue = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=["invalid_port"],  # Invalid port ID
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process],
                port_data=[queue],
                resource_data=[resource],
            )
        
        assert "port" in str(exc_info.value).lower()
        assert "R1" in str(exc_info.value)

    def test_check_resources_missing_output_port(self):
        """Test that resources without output ports raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        # Only input queue, no output queue
        queue = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=["Q1"],
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process],
                port_data=[queue],
                resource_data=[resource],
            )
        
        assert "output port" in str(exc_info.value).lower()
        assert "R1" in str(exc_info.value)

    def test_check_resources_missing_input_port(self):
        """Test that resources without input ports raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        # Only output queue, no input queue
        queue = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.OUTPUT,
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=["Q1"],
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process],
                port_data=[queue],
                resource_data=[resource],
            )
        
        assert "input port" in str(exc_info.value).lower()
        assert "R1" in str(exc_info.value)

    def test_check_products_invalid_transport_process(self):
        """Test that products with invalid transport processes raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P1": []},
            transport_process="invalid_transport",  # Invalid transport process
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process],
                product_data=[product],
            )
        
        assert "transport process" in str(exc_info.value).lower()
        assert "Product1" in str(exc_info.value)

    def test_check_products_invalid_required_process(self):
        """Test that products with invalid required processes raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P1": [], "invalid_process": []},  # Invalid process
            transport_process="TP1",
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process, transport],
                product_data=[product],
            )
        
        assert "processes" in str(exc_info.value).lower()
        assert "Product1" in str(exc_info.value)

    def test_check_sinks_invalid_product_type(self):
        """Test that sinks with invalid product types raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P1": []},
            transport_process="TP1",
        )
        
        sink = sink_data.SinkData(
            ID="Sink1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="invalid_product",  # Invalid product type
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[transport],
                product_data=[product],
                sink_data=[sink],
            )
        
        assert "product" in str(exc_info.value).lower()
        # Cross-field validation runs in fixed order; may raise on sink (Sink1) or product (processes) first
        msg = str(exc_info.value)
        assert (
            "Sink1" in msg
            or "missing or faulty" in msg.lower()
            or "processes" in msg.lower()
            or "Product1" in msg
        )

    def test_check_sinks_missing_input_port(self):
        """Test that sinks without input ports raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={},
            transport_process="TP1",
        )
        
        # Only output queue, no input queue
        queue = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[50.0, 50.0],
            interface_type=port_data.PortInterfaceType.OUTPUT,
        )
        
        sink = sink_data.SinkData(
            ID="Sink1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product1",
            ports=["Q1"],
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[transport],
                product_data=[product],
                port_data=[queue],
                sink_data=[sink],
            )
        
        assert "input port" in str(exc_info.value).lower()
        assert "Sink1" in str(exc_info.value)

    def test_check_sources_invalid_time_model(self):
        """Test that sources with invalid time models raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={},
            transport_process="TP1",
        )
        
        source = source_data.SourceData(
            ID="Source1",
            description="Source 1",
            location=[0.0, 0.0],
            product_type="Product1",
            time_model_id="invalid_tm",  # Invalid time model
            routing_heuristic="random",
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[transport],
                product_data=[product],
                source_data=[source],
            )
        
        assert "time model" in str(exc_info.value).lower()
        assert "Source1" in str(exc_info.value)

    def test_check_sources_invalid_product_type(self):
        """Test that sources with invalid product types raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={},
            transport_process="TP1",
        )
        
        source = source_data.SourceData(
            ID="Source1",
            description="Source 1",
            location=[0.0, 0.0],
            product_type="invalid_product",  # Invalid product type
            time_model_id="tm1",
            routing_heuristic="random",
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[transport],
                product_data=[product],
                source_data=[source],
            )
        
        assert "product" in str(exc_info.value).lower()
        assert "Source1" in str(exc_info.value)

    def test_check_sources_missing_output_port(self):
        """Test that sources without output ports raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={},
            transport_process="TP1",
        )
        
        # Only input queue, no output queue
        queue = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[0.0, 0.0],
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        
        source = source_data.SourceData(
            ID="Source1",
            description="Source 1",
            location=[0.0, 0.0],
            product_type="Product1",
            time_model_id="tm1",
            routing_heuristic="random",
            ports=["Q1"],
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[transport],
                product_data=[product],
                port_data=[queue],
                source_data=[source],
            )
        
        assert "output port" in str(exc_info.value).lower()
        assert "Source1" in str(exc_info.value)


class TestDuplicateIDValidation:
    """Tests for duplicate ID validation."""

    def test_duplicate_ids_across_different_types(self):
        """Test that duplicate IDs across different data types raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="shared_id",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="shared_id",  # Same ID as time model
            description="Process 1",
            time_model_id="shared_id",
            type="ProductionProcesses",
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process],
            )
        
        assert "duplicate ids" in str(exc_info.value).lower()
        assert "shared_id" in str(exc_info.value)

    def test_duplicate_ids_node_and_resource(self):
        """Test that duplicate IDs between nodes and resources raise ValidationError."""
        node = node_data.NodeData(
            ID="shared_id",
            description="Node 1",
            location=[10.0, 10.0],
        )
        
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        resource = resource_data.ResourceData(
            ID="shared_id",  # Same ID as node
            description="Resource 1",
            location=[20.0, 20.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process],
                node_data=[node],
                resource_data=[resource],
            )
        
        assert "duplicate ids" in str(exc_info.value).lower()
        assert "shared_id" in str(exc_info.value)

    def test_duplicate_ids_source_and_sink(self):
        """Test that duplicate IDs between sources and sinks raise ValidationError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={},
            transport_process="TP1",
        )
        
        source = source_data.SourceData(
            ID="shared_id",
            description="Source 1",
            location=[0.0, 0.0],
            product_type="Product1",
            time_model_id="tm1",
            routing_heuristic="random",
        )
        
        sink = sink_data.SinkData(
            ID="shared_id",  # Same ID as source
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product1",
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[transport],
                product_data=[product],
                source_data=[source],
                sink_data=[sink],
            )
        
        assert "duplicate ids" in str(exc_info.value).lower()
        assert "shared_id" in str(exc_info.value)

    def test_unique_ids_allowed(self):
        """Test that unique IDs across all types are allowed."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        node = node_data.NodeData(
            ID="N1",
            description="Node 1",
            location=[10.0, 10.0],
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[20.0, 20.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
        )
        
        # Should not raise an error
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process],
            node_data=[node],
            resource_data=[resource],
        )
        
        assert system.ID == "test"
        assert len(system.time_model_data) == 1
        assert len(system.process_data) == 1
        assert len(system.node_data) == 1
        assert len(system.resource_data) == 1


class TestIncrementalAddition:
    """Tests for adding components incrementally without validation errors."""

    def test_add_components_one_by_one(self):
        """Test that components can be added one by one without validation errors."""
        # Start with empty system
        system = ProductionSystemData(ID="test")
        
        # Add time models
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        tm2 = time_model_data.FunctionTimeModelData(
            ID="tm2",
            description="Time model 2",
            distribution_function="constant",
            location=5.0,
            scale=0.0,
        )
        system.time_model_data.append(tm1)
        system.time_model_data.append(tm2)
        system.revalidate()  # Should not raise
        
        # Add processes
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm2",
            type="TransportProcesses",
        )
        system.process_data.append(process)
        system.process_data.append(transport)
        system.revalidate()  # Should not raise
        
        # Add states
        state = state_data.BreakDownStateData(
            ID="state1",
            description="State 1",
            time_model_id="tm1",
            type="BreakDownState",
            repair_time_model_id="tm2",
        )
        system.state_data.append(state)
        system.revalidate()  # Should not raise
        
        # Add ports
        queue1 = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        queue2 = port_data.QueueData(
            ID="Q2",
            description="Queue 2",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.OUTPUT,
        )
        system.port_data.append(queue1)
        system.port_data.append(queue2)
        system.revalidate()  # Should not raise
        
        # Add resources
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            state_ids=["state1"],
            ports=["Q1", "Q2"],
        )
        system.resource_data.append(resource)
        system.revalidate()  # Should not raise
        
        # Add products
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P1": []},
            transport_process="TP1",
        )
        system.product_data.append(product)
        system.revalidate()  # Should not raise
        
        # Add sources
        source = source_data.SourceData(
            ID="Source1",
            description="Source 1",
            location=[0.0, 0.0],
            product_type="Product1",
            time_model_id="tm1",
            routing_heuristic="random",
        )
        system.source_data.append(source)
        system.revalidate()  # Should not raise
        
        # Add sinks
        sink = sink_data.SinkData(
            ID="Sink1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product1",
        )
        system.sink_data.append(sink)
        system.revalidate()  # Should not raise
        
        # Final validation
        assert len(system.time_model_data) == 2
        assert len(system.process_data) == 2
        assert len(system.state_data) == 1
        assert len(system.port_data) == 4
        assert len(system.resource_data) == 1
        assert len(system.product_data) == 1
        assert len(system.source_data) == 1
        assert len(system.sink_data) == 1

    def test_add_duplicate_id_after_initial_creation(self):
        """Test that adding duplicate IDs after initial creation is caught by revalidate."""
        system = ProductionSystemData(
            ID="test",
            time_model_data=[
                time_model_data.FunctionTimeModelData(
                    ID="tm1",
                    description="Time model 1",
                    distribution_function="constant",
                    location=10.0,
                    scale=0.0,
                )
            ],
        )
        
        # Add a process with duplicate ID
        duplicate_process = processes_data.ProductionProcessData(
            ID="tm1",  # Duplicate of time model ID
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        system.process_data.append(duplicate_process)
        
        with pytest.raises(ValidationError) as exc_info:
            system.revalidate()
        
        assert "duplicate ids" in str(exc_info.value).lower()
        assert "tm1" in str(exc_info.value)


class TestSystemTransformations:
    """Tests for system transformations like adding/removing machines."""

    @pytest.fixture
    def base_system(self):
        """Create a base valid system for transformation tests."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        queue1 = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        queue2 = port_data.QueueData(
            ID="Q2",
            description="Queue 2",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.OUTPUT,
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=["Q1", "Q2"],
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P1": []},
            transport_process="TP1",
        )
        
        source = source_data.SourceData(
            ID="Source1",
            description="Source 1",
            location=[0.0, 0.0],
            product_type="Product1",
            time_model_id="tm1",
            routing_heuristic="random",
        )
        
        sink = sink_data.SinkData(
            ID="Sink1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product1",
        )
        
        return ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process, transport],
            port_data=[queue1, queue2],
            resource_data=[resource],
            product_data=[product],
            source_data=[source],
            sink_data=[sink],
        )

    def test_add_resource_with_new_process(self, base_system):
        """Test adding a new resource with a new process."""
        # Add new time model for new process
        tm2 = time_model_data.FunctionTimeModelData(
            ID="tm2",
            description="Time model 2",
            distribution_function="constant",
            location=5.0,
            scale=0.0,
        )
        base_system.time_model_data.append(tm2)
        
        # Add new process
        process2 = processes_data.ProductionProcessData(
            ID="P2",
            description="Process 2",
            time_model_id="tm2",
            type="ProductionProcesses",
        )
        base_system.process_data.append(process2)
        
        # Add new queues
        queue3 = port_data.QueueData(
            ID="Q3",
            description="Queue 3",
            capacity=10,
            location=[20.0, 20.0],
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        queue4 = port_data.QueueData(
            ID="Q4",
            description="Queue 4",
            capacity=10,
            location=[20.0, 20.0],
            interface_type=port_data.PortInterfaceType.OUTPUT,
        )
        base_system.port_data.append(queue3)
        base_system.port_data.append(queue4)
        
        # Add new resource
        resource2 = resource_data.ResourceData(
            ID="R2",
            description="Resource 2",
            location=[20.0, 20.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P2"],
            ports=["Q3", "Q4"],
        )
        base_system.resource_data.append(resource2)
        
        # Should not raise
        base_system.revalidate()
        
        assert len(base_system.resource_data) == 2
        assert len(base_system.process_data) == 3  # P1, P2, TP1

    def test_remove_resource(self, base_system):
        """Test removing a resource."""
        # Remove resource
        base_system.resource_data.remove(base_system.resource_data[0])
        
        # Should not raise validation error (resource removal is allowed)
        base_system.revalidate()
        
        assert len(base_system.resource_data) == 0

    def test_add_resource_with_invalid_process_raises_error(self, base_system):
        """Test that adding a resource with invalid process raises ValidationError."""
        # Add new queues
        queue3 = port_data.QueueData(
            ID="Q3",
            description="Queue 3",
            capacity=10,
            location=[20.0, 20.0],
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        queue4 = port_data.QueueData(
            ID="Q4",
            description="Queue 4",
            capacity=10,
            location=[20.0, 20.0],
            interface_type=port_data.PortInterfaceType.OUTPUT,
        )
        base_system.port_data.append(queue3)
        base_system.port_data.append(queue4)
        
        # Add resource with invalid process
        resource2 = resource_data.ResourceData(
            ID="R2",
            description="Resource 2",
            location=[20.0, 20.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["invalid_process"],  # Invalid process
            ports=["Q3", "Q4"],
        )
        base_system.resource_data.append(resource2)
        
        with pytest.raises(ValidationError) as exc_info:
            base_system.revalidate()
        
        assert "process" in str(exc_info.value).lower()
        assert "R2" in str(exc_info.value)

    def test_add_resource_without_ports_gets_defaults(self, base_system):
        """Test that adding a resource without ports gets default ports."""
        # Add new process
        process2 = processes_data.ProductionProcessData(
            ID="P2",
            description="Process 2",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        base_system.process_data.append(process2)
        
        # Add resource without ports
        resource2 = resource_data.ResourceData(
            ID="R2",
            description="Resource 2",
            location=[20.0, 20.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P2"],
            ports=None,  # No ports
        )
        base_system.resource_data.append(resource2)
        
        # Add default queues
        from prodsys.models.production_system_data import add_default_queues_to_resources
        base_system = add_default_queues_to_resources(base_system)
        
        # Should not raise
        base_system.revalidate()
        
        assert resource2.ports is not None
        assert len(resource2.ports) >= 1  # At least one port

    def test_update_resource_processes(self, base_system):
        """Test updating resource processes."""
        # Add new process
        process2 = processes_data.ProductionProcessData(
            ID="P2",
            description="Process 2",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        base_system.process_data.append(process2)
        
        # Update resource to use new process
        base_system.resource_data[0].process_ids = ["P2"]
        
        # Should not raise
        base_system.revalidate()
        
        assert base_system.resource_data[0].process_ids == ["P2"]

    def test_update_resource_processes_invalid(self, base_system):
        """Test that updating resource with invalid process raises ValidationError."""
        # Update resource to use invalid process
        base_system.resource_data[0].process_ids = ["invalid_process"]
        
        with pytest.raises(ValidationError) as exc_info:
            base_system.revalidate()
        
        assert "process" in str(exc_info.value).lower()


class TestConfigurationValidation:
    """Tests for configuration validation methods."""

    def test_assert_no_redundant_locations(self):
        """Test that redundant locations raise ValueError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        # Two resources at same location
        resource1 = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
        )
        
        resource2 = resource_data.ResourceData(
            ID="R2",
            description="Resource 2",
            location=[10.0, 10.0],  # Same location
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process],
            resource_data=[resource1, resource2],
        )
        
        with pytest.raises(ValueError) as exc_info:
            system.validate_configuration()
        
        assert "location" in str(exc_info.value).lower()

    def test_assert_all_links_available(self):
        """Test that invalid links raise ValueError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        # Link transport process with invalid link
        link_transport = processes_data.LinkTransportProcessData(
            ID="LTP1",
            description="Link Transport 1",
            time_model_id="tm1",
            type="LinkTransportProcesses",
            links=[["invalid_location", "R1"]],  # Invalid start location
        )
        
        node = node_data.NodeData(
            ID="N1",
            description="Node 1",
            location=[10.0, 10.0],
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[20.0, 20.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["LTP1"],
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[link_transport],
            node_data=[node],
            resource_data=[resource],
        )
        
        with pytest.raises(ValueError) as exc_info:
            system.validate_configuration()
        
        assert "link" in str(exc_info.value).lower()
        assert "invalid_location" in str(exc_info.value)

    def test_assert_ports_have_locations(self):
        """Test that ports without locations raise ValueError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        # Port without location
        queue = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=None,  # No location
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        
        # Output port with location (required for resource)
        queue2 = port_data.QueueData(
            ID="Q2",
            description="Queue 2",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.OUTPUT,
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=["Q1", "Q2"],
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process],
            port_data=[queue, queue2],
            resource_data=[resource],
        )
        
        with pytest.raises(ValueError) as exc_info:
            system.validate_configuration()
        
        assert "location" in str(exc_info.value).lower()
        assert "Q1" in str(exc_info.value)

    def test_assert_required_processes_in_resources_available(self):
        """Test that missing required processes raise ValueError."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        # Add P2 process but don't assign it to any resource
        process2 = processes_data.ProductionProcessData(
            ID="P2",
            description="Process 2",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        # Product requires P2, but no resource provides it
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P2": []},  # P2 not available in resources
            transport_process="TP1",
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],  # Only P1, not P2
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process, process2, transport],
            resource_data=[resource],
            product_data=[product],
        )
        
        with pytest.raises(ValueError) as exc_info:
            system.validate_configuration()
        
        assert "process" in str(exc_info.value).lower()
        assert "P2" in str(exc_info.value) or "not available" in str(exc_info.value).lower()

    def test_valid_configuration_passes(self):
        """Test that a valid configuration passes all validations."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        queue1 = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        queue2 = port_data.QueueData(
            ID="Q2",
            description="Queue 2",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.OUTPUT,
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=["Q1", "Q2"],
        )
        
        # Add transport resource for TP1
        transport_resource = resource_data.ResourceData(
            ID="TR1",
            description="Transport Resource 1",
            location=[5.0, 5.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["TP1"],
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P1": []},
            transport_process="TP1",
        )
        
        source = source_data.SourceData(
            ID="Source1",
            description="Source 1",
            location=[0.0, 0.0],
            product_type="Product1",
            time_model_id="tm1",
            routing_heuristic="random",
        )
        
        sink = sink_data.SinkData(
            ID="Sink1",
            description="Sink 1",
            location=[50.0, 50.0],
            product_type="Product1",
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process, transport],
            port_data=[queue1, queue2],
            resource_data=[resource, transport_resource],
            product_data=[product],
            source_data=[source],
            sink_data=[sink],
        )
        
        # Should not raise
        system.validate_configuration()
        
        assert system.valid_configuration is True


class TestScheduleValidation:
    """Tests for schedule validation."""

    def test_check_schedule_invalid_resource(self):
        """Test that schedule with invalid resource raises ValidationError."""
        from prodsys.models.performance_data import Event
        
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P1": []},
            transport_process="TP1",
        )
        
        # Event with invalid resource
        event = Event(
            resource="invalid_resource",
            product="Product1_1",
            process="P1",
            activity="start state",
            state_type="Production",
            time=0.0,
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process, transport],
                resource_data=[resource],
                product_data=[product],
                schedule=[event],
            )
        
        assert "resource" in str(exc_info.value).lower()
        assert "schedule" in str(exc_info.value).lower()

    def test_check_schedule_invalid_process(self):
        """Test that schedule with invalid process raises ValidationError."""
        from prodsys.models.performance_data import Event
        
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P1": []},
            transport_process="TP1",
        )
        
        # Event with invalid process
        event = Event(
            resource="R1",
            product="Product1_1",
            process="invalid_process",
            activity="start state",
            state_type="Production",
            time=0.0,
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process, transport],
                resource_data=[resource],
                product_data=[product],
                schedule=[event],
            )
        
        assert "process" in str(exc_info.value).lower()
        assert "schedule" in str(exc_info.value).lower()

    def test_check_schedule_invalid_product(self):
        """Test that schedule with invalid product raises ValidationError."""
        from prodsys.models.performance_data import Event
        
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
        )
        
        product = product_data.ProductData(
            ID="Product1",
            description="Product 1",
            type="Product1",
            processes={"P1": []},
            transport_process="TP1",
        )
        
        # Event with invalid product (Product2 doesn't exist)
        event = Event(
            resource="R1",
            product="Product2_1",  # Invalid product
            process="P1",
            activity="start state",
            state_type="Production",
            time=0.0,
        )
        
        with pytest.raises(ValidationError) as exc_info:
            ProductionSystemData(
                ID="test",
                time_model_data=[tm1],
                process_data=[process, transport],
                resource_data=[resource],
                product_data=[product],
                schedule=[event],
            )
        
        assert "product" in str(exc_info.value).lower()
        assert "schedule" in str(exc_info.value).lower()


class TestDefaultQueueCreation:
    """Tests for automatic default queue creation for resources."""
    
    def test_production_resource_gets_default_queues(self):
        """Test that a production resource without ports gets default queues automatically."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=None,  # No ports specified
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process],
            resource_data=[resource],
        )
        
        # Resource should have default queues added
        assert resource.ports is not None
        assert len(resource.ports) == 1
        assert f"{resource.ID}_default_queue" in resource.ports
        
        # Queues should be in port_data
        port_ids = [p.ID for p in system.port_data]
        assert f"{resource.ID}_default_queue" in port_ids
    
    def test_transport_resource_no_queues(self):
        """Test that a transport resource with only transport processes doesn't get queues."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        resource = resource_data.ResourceData(
            ID="TR1",
            description="Transport Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["TP1"],
            ports=None,  # No ports specified
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[transport],
            resource_data=[resource],
        )
        
        # Transport resource should not have queues added automatically
        assert resource.ports is None or len(resource.ports) == 0
        
        # No default queues should be in port_data for this resource
        port_ids = [p.ID for p in system.port_data]
        assert f"{resource.ID}_default_input_queue" not in port_ids
        assert f"{resource.ID}_default_output_queue" not in port_ids
        assert f"{resource.ID}_default_queue" not in port_ids
    
    def test_mixed_process_resource_gets_queues(self):
        """Test that a resource with both production and transport processes gets queues."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        transport = processes_data.TransportProcessData(
            ID="TP1",
            description="Transport 1",
            time_model_id="tm1",
            type="TransportProcesses",
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1", "TP1"],  # Both production and transport
            ports=None,  # No ports specified
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process, transport],
            resource_data=[resource],
        )
        
        # Resource should have default queues added (has production process)
        assert resource.ports is not None
        assert len(resource.ports) == 1
        assert f"{resource.ID}_default_queue" in resource.ports
    
    def test_resource_with_existing_ports_no_auto_creation(self):
        """Test that a resource with existing ports doesn't get additional default queues."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        queue1 = port_data.QueueData(
            ID="Q1",
            description="Queue 1",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.INPUT,
        )
        queue2 = port_data.QueueData(
            ID="Q2",
            description="Queue 2",
            capacity=10,
            location=[10.0, 10.0],
            interface_type=port_data.PortInterfaceType.OUTPUT,
        )
        
        resource = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=["Q1", "Q2"],  # Ports already specified
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process],
            port_data=[queue1, queue2],
            resource_data=[resource],
        )
        
        # Resource should keep its existing ports
        assert resource.ports == ["Q1", "Q2"]
        
        # No default queues should be created
        port_ids = [p.ID for p in system.port_data]
        assert f"{resource.ID}_default_input_queue" not in port_ids
        assert f"{resource.ID}_default_output_queue" not in port_ids
    
    def test_multiple_resources_get_default_queues(self):
        """Test that multiple production resources without ports all get default queues."""
        tm1 = time_model_data.FunctionTimeModelData(
            ID="tm1",
            description="Time model 1",
            distribution_function="constant",
            location=10.0,
            scale=0.0,
        )
        
        process = processes_data.ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type="ProductionProcesses",
        )
        
        resource1 = resource_data.ResourceData(
            ID="R1",
            description="Resource 1",
            location=[10.0, 10.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=None,
        )
        
        resource2 = resource_data.ResourceData(
            ID="R2",
            description="Resource 2",
            location=[20.0, 20.0],
            capacity=1,
            controller="PipelineController",
            control_policy="FIFO",
            process_ids=["P1"],
            ports=None,
        )
        
        system = ProductionSystemData(
            ID="test",
            time_model_data=[tm1],
            process_data=[process],
            resource_data=[resource1, resource2],
        )
        
        # Both resources should have default queues
        assert resource1.ports is not None
        assert len(resource1.ports) == 1
        assert resource2.ports is not None
        assert len(resource2.ports) == 1
        
        # All queues should be in port_data
        port_ids = [p.ID for p in system.port_data]
        assert f"{resource1.ID}_default_queue" in port_ids
        assert f"{resource2.ID}_default_queue" in port_ids


class TestDependencySerialization:
    """Tests for dependency data serialization and deserialization."""

    def test_dependency_serialization_deserialization(self):
        """Test that dependency data is correctly serialized and deserialized."""
        import json
        import tempfile
        import os
        import prodsys.express as psx
        from prodsys.models.dependency_data import ProcessDependencyData, ResourceDependencyData

        # Create time models
        t1 = psx.FunctionTimeModel("normal", 1, 0.1, "t1")
        t2 = psx.FunctionTimeModel("normal", 2, 0.2, "t2")
        t3 = psx.DistanceTimeModel(speed=180, reaction_time=0.1, ID="t3")

        # Create processes
        p1 = psx.ProductionProcess(t1, "p1")
        p2 = psx.ProductionProcess(t2, "p2")
        tp = psx.TransportProcess(t3, "tp")
        move_p = psx.TransportProcess(t3, "move")
        assembly_process = psx.ProductionProcess(
            psx.FunctionTimeModel("exponential", 0.1, ID="assembly_time"),
            "assembly_process"
        )

        # Create setup states
        s1 = psx.FunctionTimeModel("exponential", 0.5, ID="s1")
        setup_state_1 = psx.SetupState(s1, p1, p2, "S1")
        setup_state_2 = psx.SetupState(s1, p2, p1, "S2")

        # Create resources
        worker = psx.Resource([move_p, assembly_process], [2, 0], 1, ID="worker")
        worker2 = psx.Resource([move_p, assembly_process], [3, 0], 1, ID="worker2")
        transport = psx.Resource([tp], [2, 2], 1, ID="transport")

        # Create interaction nodes
        interaction_node_assembly = psx.Node(location=[5, 6], ID="interaction_node_assembly")
        interaction_node_resource_2 = psx.Node(location=[7, 4], ID="interaction_node_resource_2")

        # Create dependencies
        assembly_dependency = psx.ProcessDependency(
            ID="assembly_dependency",
            required_process=assembly_process,
            interaction_node=interaction_node_assembly,
        )
        resource_2_dependency = psx.ResourceDependency(
            ID="resource_2_dependency",
            required_resource=worker2,
            interaction_node=interaction_node_resource_2,
        )

        # Create resources with dependencies
        machine = psx.Resource(
            [p1, p2],
            [5, 5],
            1,
            states=[setup_state_1, setup_state_2],
            ID="machine",
            dependencies=[assembly_dependency],
        )
        machine2 = psx.Resource(
            [p1, p2],
            [7, 2],
            3,
            states=[setup_state_1, setup_state_2],
            ID="machine2",
            dependencies=[resource_2_dependency],
        )

        # Create products
        product1 = psx.Product(process=[p1], transport_process=tp, ID="product1")
        product2 = psx.Product(process=[p2], transport_process=tp, ID="product2")

        # Create sources and sinks
        arrival_model_1 = psx.FunctionTimeModel("exponential", 2, ID="arrival_model_1")
        arrival_model_2 = psx.FunctionTimeModel("exponential", 4, ID="arrival_model_2")
        source1 = psx.Source(product1, arrival_model_1, [0, 0], ID="source_1")
        source2 = psx.Source(product2, arrival_model_2, [0, 0], ID="source_2")
        sink1 = psx.Sink(product1, [10, 0], "sink1")
        sink2 = psx.Sink(product2, [10, 0], "sink2")

        # Create production system
        system = psx.ProductionSystem(
            [machine, machine2, transport, worker, worker2],
            [source1, source2],
            [sink1, sink2],
        )

        # Convert to ProductionSystemData model
        model = system.to_model()

        # Verify original dependency data
        assert len(model.dependency_data) == 2
        original_dependency_ids = {dep.ID for dep in model.dependency_data}
        assert "assembly_dependency" in original_dependency_ids
        assert "resource_2_dependency" in original_dependency_ids

        # Check that dependencies are present in resource data
        machine_resource = next(r for r in model.resource_data if r.ID == "machine")
        assert machine_resource.dependency_ids == ["assembly_dependency"]
        machine2_resource = next(r for r in model.resource_data if r.ID == "machine2")
        assert machine2_resource.dependency_ids == ["resource_2_dependency"]

        # Serialize to JSON
        json_str = model.model_dump_json(indent=2)
        json_dict = json.loads(json_str)
        assert len(json_dict.get("dependency_data", [])) == 2

        # Deserialize from JSON
        deserialized_model = ProductionSystemData.model_validate(json_dict)

        # Verify deserialized dependency data
        assert len(deserialized_model.dependency_data) == 2
        deserialized_dependency_ids = {dep.ID for dep in deserialized_model.dependency_data}
        assert deserialized_dependency_ids == original_dependency_ids

        # Verify each dependency in detail
        for orig_dep in model.dependency_data:
            deser_dep = next(
                (d for d in deserialized_model.dependency_data if d.ID == orig_dep.ID),
                None
            )
            assert deser_dep is not None, f"Dependency '{orig_dep.ID}' not found in deserialized data"
            assert orig_dep.dependency_type == deser_dep.dependency_type

            # Verify dependency-specific fields
            if isinstance(orig_dep, ProcessDependencyData):
                assert orig_dep.required_process == deser_dep.required_process
                assert orig_dep.interaction_node == deser_dep.interaction_node
            elif isinstance(orig_dep, ResourceDependencyData):
                assert orig_dep.required_resource == deser_dep.required_resource
                assert orig_dep.interaction_node == deser_dep.interaction_node

        # Verify resource dependency references are preserved
        deserialized_machine = next(r for r in deserialized_model.resource_data if r.ID == "machine")
        assert deserialized_machine.dependency_ids == machine_resource.dependency_ids
        deserialized_machine2 = next(r for r in deserialized_model.resource_data if r.ID == "machine2")
        assert deserialized_machine2.dependency_ids == machine2_resource.dependency_ids

    def test_dependency_file_io(self):
        """Test that dependency data is correctly preserved when writing to and reading from file."""
        import tempfile
        import os
        import prodsys.express as psx

        # Create a simple system with dependencies
        t1 = psx.FunctionTimeModel("normal", 1, 0.1, "t1")
        p1 = psx.ProductionProcess(t1, "p1")
        tp = psx.TransportProcess(psx.DistanceTimeModel(speed=180, reaction_time=0.1, ID="t3"), "tp")
        assembly_process = psx.ProductionProcess(
            psx.FunctionTimeModel("exponential", 0.1, ID="assembly_time"),
            "assembly_process"
        )
        worker = psx.Resource([assembly_process], [2, 0], 1, ID="worker")
        interaction_node = psx.Node(location=[5, 6], ID="interaction_node_assembly")
        assembly_dependency = psx.ProcessDependency(
            ID="assembly_dependency",
            required_process=assembly_process,
            interaction_node=interaction_node,
        )
        machine = psx.Resource(
            [p1],
            [5, 5],
            1,
            ID="machine",
            dependencies=[assembly_dependency],
        )
        transport = psx.Resource([tp], [2, 2], 1, ID="transport")
        product1 = psx.Product(process=[p1], transport_process=tp, ID="product1")
        source1 = psx.Source(
            product1,
            psx.FunctionTimeModel("exponential", 2, ID="arrival_model_1"),
            [0, 0],
            ID="source_1"
        )
        sink1 = psx.Sink(product1, [10, 0], "sink1")

        system = psx.ProductionSystem(
            [machine, transport, worker],
            [source1],
            [sink1],
        )

        model = system.to_model()
        original_dependency_count = len(model.dependency_data)

        # Write to file
        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as tmp_file:
            tmp_path = tmp_file.name

        try:
            model.write(tmp_path)

            # Read from file
            loaded_model = ProductionSystemData.read(tmp_path)

            # Verify loaded dependency data
            assert len(loaded_model.dependency_data) == original_dependency_count
            loaded_dependency_ids = {dep.ID for dep in loaded_model.dependency_data}
            original_dependency_ids = {dep.ID for dep in model.dependency_data}
            assert loaded_dependency_ids == original_dependency_ids

            # Verify resource dependencies are preserved
            for orig_resource in model.resource_data:
                if orig_resource.dependency_ids:
                    loaded_resource = next(
                        (r for r in loaded_model.resource_data if r.ID == orig_resource.ID),
                        None
                    )
                    assert loaded_resource is not None
                    assert orig_resource.dependency_ids == loaded_resource.dependency_ids

        finally:
            if os.path.exists(tmp_path):
                os.unlink(tmp_path)

    def test_none_dependency_data_handling(self):
        """Test that None dependency_data is handled correctly."""
        import prodsys.express as psx

        # Create a simple system
        t1 = psx.FunctionTimeModel("normal", 1, 0.1, "t1")
        p1 = psx.ProductionProcess(t1, "p1")
        tp = psx.TransportProcess(psx.DistanceTimeModel(speed=180, reaction_time=0.1, ID="t3"), "tp")
        transport = psx.Resource([tp], [2, 2], 1, ID="transport")
        machine = psx.Resource([p1], [5, 5], 1, ID="machine")  # Resource that provides p1
        product1 = psx.Product(process=[p1], transport_process=tp, ID="product1")
        source1 = psx.Source(
            product1,
            psx.FunctionTimeModel("exponential", 2, ID="arrival_model_1"),
            [0, 0],
            ID="source_1"
        )
        sink1 = psx.Sink(product1, [10, 0], "sink1")

        system = psx.ProductionSystem([machine, transport], [source1], [sink1])
        model = system.to_model()

        # Test with None dependency_data
        model_dict = model.model_dump()
        model_dict['dependency_data'] = None
        model_with_none = ProductionSystemData.model_validate(model_dict)

        # Should default to empty list or None
        assert model_with_none.dependency_data is None or len(model_with_none.dependency_data) == 0




#!/usr/bin/env python3
"""
Simple example demonstrating ProcessModel and SystemResource features.

This script shows the basic functionality without complex dependencies.
"""

import prodsys.express as psx

def main():
    """Main example function."""
    print("=== Simple ProcessModel and SystemResource Example ===\n")
    
    # 1. Create time models
    print("1. Creating time models...")
    tm_weld = psx.FunctionTimeModel("normal", 20.0, 5.0, "tm_weld")
    tm_assembly = psx.FunctionTimeModel("normal", 15.0, 3.0, "tm_assembly")
    tm_transport = psx.DistanceTimeModel(60, 0.1, "manhattan", "tm_transport")
    print("   ✓ Time models created")
    
    # 2. Create individual processes
    print("\n2. Creating individual processes...")
    weld_process = psx.ProductionProcess(tm_weld, "weld")
    assembly_process = psx.ProductionProcess(tm_assembly, "assembly")
    transport_process = psx.TransportProcess(tm_transport, "transport")
    print("   ✓ Individual processes created")
    
    # 3. Create SequentialProcess (auto-generates adjacency matrix)
    print("\n3. Creating SequentialProcess...")
    sequential_process = psx.SequentialProcess(
        time_model=psx.FunctionTimeModel("constant", 0.0, ID="tm_seq"),
        process_ids=["weld", "assembly"],
        ID="sequential_process"
    )
    print(f"   ✓ SequentialProcess created")
    print(f"   ✓ Adjacency matrix: {sequential_process.adjacency_matrix}")
    
    # 4. Create ProcessModel (DAG)
    print("\n4. Creating ProcessModel (DAG)...")
    process_model = psx.ProcessModel(
        time_model=psx.FunctionTimeModel("constant", 0.0, ID="tm_pm"),
        process_ids=["weld", "assembly"],
        adjacency_matrix={
            "weld": ["assembly"],
            "assembly": []
        },
        ID="process_model"
    )
    print("   ✓ ProcessModel created")
    
    # 5. Create LoadingProcess
    print("\n5. Creating LoadingProcess...")
    loading_process = psx.LoadingProcess(
        time_model=psx.FunctionTimeModel("normal", 2.0, 0.5, ID="tm_loading"),
        dependency_type="before",
        can_be_chained=True,
        ID="loading_process"
    )
    print("   ✓ LoadingProcess created")
    
    # 6. Create regular resources
    print("\n6. Creating regular resources...")
    machine1 = psx.Resource([weld_process], [5, 0], 1, ID="machine1")
    machine2 = psx.Resource([assembly_process], [10, 0], 1, ID="machine2")
    transport_resource = psx.Resource([transport_process], [0, 0], 1, ID="transport")
    print("   ✓ Regular resources created")
    
    # 7. Create SystemResource (simplified)
    print("\n7. Creating SystemResource...")
    system_resource = psx.SystemResource(
        processes=[sequential_process],
        location=[20, 5],
        subresource_ids=["machine1", "machine2"],
        ID="system_resource"
    )
    print("   ✓ SystemResource created")
    
    # 8. Create Product with ProcessModel
    print("\n8. Creating Product with ProcessModel...")
    product = psx.Product(
        process=sequential_process,
        transport_process=transport_process,
        ID="product"
    )
    print("   ✓ Product created")
    
    # 9. Test data model conversion
    print("\n9. Testing data model conversion...")
    try:
        # Convert to data models
        seq_data = sequential_process.to_model()
        pm_data = process_model.to_model()
        lp_data = loading_process.to_model()
        sr_data = system_resource.to_model()
        product_data = product.to_model()
        
        print("   ✓ All objects converted to data models")
        print(f"   ✓ SequentialProcess data: ID={seq_data.ID}, process_ids={seq_data.process_ids}")
        print(f"   ✓ ProcessModel data: ID={pm_data.ID}, process_ids={pm_data.process_ids}")
        print(f"   ✓ LoadingProcess data: ID={lp_data.ID}, dependency_type={lp_data.dependency_type}")
        print(f"   ✓ SystemResource data: ID={sr_data.ID}, subresource_ids={sr_data.subresource_ids}")
        print(f"   ✓ Product data: ID={product_data.ID}, processes={product_data.processes}")
        
    except Exception as e:
        print(f"   ✗ Error converting to data model: {e}")
    
    print("\n=== Example completed successfully! ===")
    print("\nKey features demonstrated:")
    print("- ProcessModel: DAG-based process execution")
    print("- SequentialProcess: Auto-generated sequential execution")
    print("- LoadingProcess: Dependency-based loading")
    print("- SystemResource: Hierarchical resource with subresources")
    print("- Product: Uses ProcessModel as its process")

if __name__ == "__main__":
    main()



"""
Tests for order source functionality.
"""

import pytest
import prodsys.express as psx
from prodsys.models import production_system_data
from prodsys.simulation import runner


def test_order_source_creation():
    """Test that OrderSource can be created."""
    t1 = psx.FunctionTimeModel("constant", 5.0, 0, "t1")
    p1 = psx.ProductionProcess(t1, "P1")
    
    t_transport = psx.FunctionTimeModel("constant", 0.5, 0, ID="t_transport")
    tp = psx.TransportProcess(t_transport, "TP")
    
    resource = psx.Resource([p1], [10, 0], 1, ID="R1")
    transport = psx.Resource([tp], [5, 0], 1, ID="AGV1")
    
    product = psx.Product([p1], tp, "Product_A")
    
    sink = psx.Sink(product, [20, 0], "Sink")
    
    # Create orders
    order1 = psx.Order(
        ID="order1",
        ordered_products=[psx.OrderedProduct(product=product, quantity=2)],
        order_time=0.0,
        release_time=10.0,
        priority=1,
    )
    
    order_source = psx.OrderSource(
        orders=[order1],
        location=[0, 0],
        ID="OrderSource",
    )
    
    system = psx.ProductionSystem([resource, transport], [order_source], [sink])
    adapter = system.to_model()
    
    assert len(adapter.source_data) == 1
    from prodsys.models.source_data import OrderSourceData
    assert isinstance(adapter.source_data[0], OrderSourceData)
    assert adapter.source_data[0].ID == "OrderSource"
    assert len(adapter.order_data) == 1
    assert adapter.order_data[0].ID == "order1"


def test_order_source_simulation():
    """Test that simulation runs with OrderSource."""
    t1 = psx.FunctionTimeModel("constant", 5.0, 0, "t1")
    p1 = psx.ProductionProcess(t1, "P1")
    
    t_transport = psx.FunctionTimeModel("constant", 0.5, 0, ID="t_transport")
    tp = psx.TransportProcess(t_transport, "TP")
    
    resource = psx.Resource([p1], [10, 0], 1, ID="R1")
    transport = psx.Resource([tp], [5, 0], 1, ID="AGV1")
    
    product = psx.Product([p1], tp, "Product_A")
    
    sink = psx.Sink(product, [20, 0], "Sink")
    
    # Create orders with different release times
    order1 = psx.Order(
        ID="order1",
        ordered_products=[psx.OrderedProduct(product=product, quantity=2)],
        order_time=0.0,
        release_time=10.0,
        priority=1,
    )
    
    order2 = psx.Order(
        ID="order2",
        ordered_products=[psx.OrderedProduct(product=product, quantity=1)],
        order_time=5.0,
        release_time=20.0,
        priority=1,
    )
    
    order_source = psx.OrderSource(
        orders=[order1, order2],
        location=[0, 0],
        ID="OrderSource",
    )
    
    system = psx.ProductionSystem([resource, transport], [order_source], [sink])
    adapter = system.to_model()
    
    runner_instance = runner.Runner(production_system_data=adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(100)
    
    assert runner_instance.env.now == 100
    
    # Get performance data
    performance = runner_instance.get_performance_data()
    assert performance.event_log is not None
    assert len(performance.event_log) > 0


def test_order_source_multiple_product_types():
    """Test that OrderSource can release products of different types."""
    t1 = psx.FunctionTimeModel("constant", 5.0, 0, "t1")
    p1 = psx.ProductionProcess(t1, "P1")
    
    t_transport = psx.FunctionTimeModel("constant", 0.5, 0, ID="t_transport")
    tp = psx.TransportProcess(t_transport, "TP")
    
    resource = psx.Resource([p1], [10, 0], 1, ID="R1")
    transport = psx.Resource([tp], [5, 0], 1, ID="AGV1")
    
    product1 = psx.Product([p1], tp, "Product_A")
    product2 = psx.Product([p1], tp, "Product_B")
    
    sink1 = psx.Sink(product1, [20, 0], "Sink1")
    sink2 = psx.Sink(product2, [20, 0], "Sink2")
    
    # Create order with multiple product types
    order1 = psx.Order(
        ID="order1",
        ordered_products=[
            psx.OrderedProduct(product=product1, quantity=1),
            psx.OrderedProduct(product=product2, quantity=1),
        ],
        order_time=0.0,
        release_time=10.0,
        priority=1,
    )
    
    order_source = psx.OrderSource(
        orders=[order1],
        location=[0, 0],
        ID="OrderSource",
    )
    
    system = psx.ProductionSystem(
        [resource, transport], [order_source], [sink1, sink2]
    )
    adapter = system.to_model()
    
    runner_instance = runner.Runner(production_system_data=adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(100)
    
    assert runner_instance.env.now == 100
    
    # Get performance data
    performance = runner_instance.get_performance_data()
    assert performance.event_log is not None
    assert len(performance.event_log) > 0


def test_order_source_with_conwip():
    """Test that OrderSource respects ConWip limits."""
    t1 = psx.FunctionTimeModel("constant", 5.0, 0, "t1")
    p1 = psx.ProductionProcess(t1, "P1")
    
    t_transport = psx.FunctionTimeModel("constant", 0.5, 0, ID="t_transport")
    tp = psx.TransportProcess(t_transport, "TP")
    
    resource = psx.Resource([p1], [10, 0], 1, ID="R1")
    transport = psx.Resource([tp], [5, 0], 1, ID="AGV1")
    
    product = psx.Product([p1], tp, "Product_A")
    
    sink = psx.Sink(product, [20, 0], "Sink")
    
    # Create orders
    order1 = psx.Order(
        ID="order1",
        ordered_products=[psx.OrderedProduct(product=product, quantity=5)],
        order_time=0.0,
        release_time=10.0,
        priority=1,
    )
    
    order_source = psx.OrderSource(
        orders=[order1],
        location=[0, 0],
        ID="OrderSource",
    )
    
    system = psx.ProductionSystem([resource, transport], [order_source], [sink])
    adapter = system.to_model()
    adapter.conwip_number = 3
    
    runner_instance = runner.Runner(production_system_data=adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(100)
    
    assert runner_instance.env.now == 100
    
    # Get performance data
    performance = runner_instance.get_performance_data()
    assert performance.event_log is not None
    assert len(performance.event_log) > 0
    
    # Check that ConWip limit is respected
    # The number of products in the system should not exceed ConWip limit
    # (allowing for products that are being processed)
    wip_kpis = [kpi for kpi in performance.kpis if kpi.name.value == "WIP"]
    if wip_kpis:
        # WIP might temporarily exceed ConWip due to processing, but should be controlled
        assert wip_kpis[0].value <= adapter.conwip_number + 2  # Allow some margin for processing


def test_order_source_release_time():
    """Test that OrderSource releases products at the correct release time."""
    t1 = psx.FunctionTimeModel("constant", 5.0, 0, "t1")
    p1 = psx.ProductionProcess(t1, "P1")
    
    t_transport = psx.FunctionTimeModel("constant", 0.5, 0, ID="t_transport")
    tp = psx.TransportProcess(t_transport, "TP")
    
    resource = psx.Resource([p1], [10, 0], 1, ID="R1")
    transport = psx.Resource([tp], [5, 0], 1, ID="AGV1")
    
    product = psx.Product([p1], tp, "Product_A")
    
    sink = psx.Sink(product, [20, 0], "Sink")
    
    # Create order with release time
    order1 = psx.Order(
        ID="order1",
        ordered_products=[psx.OrderedProduct(product=product, quantity=1)],
        order_time=0.0,
        release_time=50.0,  # Release at time 50
        priority=1,
    )
    
    order_source = psx.OrderSource(
        orders=[order1],
        location=[0, 0],
        ID="OrderSource",
    )
    
    system = psx.ProductionSystem([resource, transport], [order_source], [sink])
    adapter = system.to_model()
    
    runner_instance = runner.Runner(production_system_data=adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(100)
    
    assert runner_instance.env.now == 100
    
    # Get performance data
    performance = runner_instance.get_performance_data()
    assert performance.event_log is not None
    
    # Check that products are released at or after release_time
    source_events = [
        e for e in performance.event_log
        if e.resource == "OrderSource" and e.activity == "start state"
    ]
    if source_events:
        # Products should be released at or after release_time (50.0)
        for event in source_events:
            assert event.time >= 50.0


def test_order_source_without_release_time():
    """Test that OrderSource uses order_time when release_time is not specified."""
    t1 = psx.FunctionTimeModel("constant", 5.0, 0, "t1")
    p1 = psx.ProductionProcess(t1, "P1")
    
    t_transport = psx.FunctionTimeModel("constant", 0.5, 0, ID="t_transport")
    tp = psx.TransportProcess(t_transport, "TP")
    
    resource = psx.Resource([p1], [10, 0], 1, ID="R1")
    transport = psx.Resource([tp], [5, 0], 1, ID="AGV1")
    
    product = psx.Product([p1], tp, "Product_A")
    
    sink = psx.Sink(product, [20, 0], "Sink")
    
    # Create order without release_time (should use order_time)
    order1 = psx.Order(
        ID="order1",
        ordered_products=[psx.OrderedProduct(product=product, quantity=1)],
        order_time=10.0,  # Should be used as release_time
        release_time=None,
        priority=1,
    )
    
    order_source = psx.OrderSource(
        orders=[order1],
        location=[0, 0],
        ID="OrderSource",
    )
    
    system = psx.ProductionSystem([resource, transport], [order_source], [sink])
    adapter = system.to_model()
    
    runner_instance = runner.Runner(production_system_data=adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(100)
    
    assert runner_instance.env.now == 100
    
    # Get performance data
    performance = runner_instance.get_performance_data()
    assert performance.event_log is not None
    assert len(performance.event_log) > 0




"""
Tests for processes_data module.
"""

import pytest
from prodsys.models.processes_data import (
    ProcessData,
    ProductionProcessData,
    CapabilityProcessData,
    TransportProcessData,
    ReworkProcessData,
    CompoundProcessData,
    RequiredCapabilityProcessData,
    LinkTransportProcessData,
    ProcessModelData,
    LoadingProcessData,
    ProcessTypeEnum,
)
from prodsys.models.time_model_data import FunctionTimeModelData
from prodsys.models.production_system_data import ProductionSystemData
from prodsys.util.statistical_functions import FunctionTimeModelEnum


@pytest.fixture
def sample_time_model():
    """Create a sample time model for testing."""
    return FunctionTimeModelData(
        ID="tm1",
        description="Test time model",
        distribution_function=FunctionTimeModelEnum.Normal,
        location=20.0,
        scale=5.0,
    )


@pytest.fixture
def sample_production_system(sample_time_model):
    """Create a minimal production system for testing hash methods."""
    from prodsys.models.production_system_data import ProductionSystemData
    
    # Create a minimal adapter with time models
    return ProductionSystemData(
        time_model_data=[sample_time_model],
        process_data=[],
        resource_data=[],
        product_data=[],
        source_data=[],
        sink_data=[],
        state_data=[],
        port_data=[],
        primitive_data=[],
        dependency_data=[],
        node_data=[],
        scenario_data=None,
    )


class TestProductionProcessData:
    """Tests for ProductionProcessData."""

    def test_creation_with_required_fields(self):
        """Test creating ProductionProcessData with required fields."""
        process = ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.ProductionProcesses,
        )
        assert process.ID == "P1"
        assert process.description == "Process 1"
        assert process.time_model_id == "tm1"
        assert process.type == ProcessTypeEnum.ProductionProcesses

    def test_default_failure_rate(self):
        """Test that failure_rate defaults to 0.0."""
        process = ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.ProductionProcesses,
        )
        assert process.failure_rate == 0.0

    def test_with_failure_rate(self):
        """Test creating ProductionProcessData with failure rate."""
        process = ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.ProductionProcesses,
            failure_rate=0.1,
        )
        assert process.failure_rate == 0.1

    def test_default_dependency_ids(self):
        """Test that dependency_ids defaults to empty list."""
        process = ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.ProductionProcesses,
        )
        assert process.dependency_ids == []

    def test_with_dependency_ids(self):
        """Test creating ProductionProcessData with dependency IDs."""
        process = ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.ProductionProcesses,
            dependency_ids=["dep1", "dep2"],
        )
        assert process.dependency_ids == ["dep1", "dep2"]

    def test_hash(self, sample_production_system, sample_time_model):
        """Test hash method."""
        process1 = ProductionProcessData(
            ID="P1",
            description="Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.ProductionProcesses,
            failure_rate=0.1,
        )
        process2 = ProductionProcessData(
            ID="P2",
            description="Process 2",
            time_model_id="tm1",
            type=ProcessTypeEnum.ProductionProcesses,
            failure_rate=0.1,
        )
        process3 = ProductionProcessData(
            ID="P3",
            description="Process 3",
            time_model_id="tm1",
            type=ProcessTypeEnum.ProductionProcesses,
            failure_rate=0.2,  # different failure rate
        )
        
        hash1 = process1.hash(sample_production_system)
        hash2 = process2.hash(sample_production_system)
        hash3 = process3.hash(sample_production_system)
        
        # Same parameters should produce same hash
        assert hash1 == hash2
        # Different failure rate should produce different hash
        assert hash1 != hash3


class TestCapabilityProcessData:
    """Tests for CapabilityProcessData."""

    def test_creation_with_capability(self):
        """Test creating CapabilityProcessData with capability."""
        process = CapabilityProcessData(
            ID="CP1",
            description="Capability Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.CapabilityProcesses,
            capability="C1",
        )
        assert process.ID == "CP1"
        assert process.capability == "C1"
        assert process.type == ProcessTypeEnum.CapabilityProcesses

    def test_hash(self, sample_production_system, sample_time_model):
        """Test hash method considers capability."""
        process1 = CapabilityProcessData(
            ID="CP1",
            description="Capability Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.CapabilityProcesses,
            capability="C1",
        )
        process2 = CapabilityProcessData(
            ID="CP2",
            description="Capability Process 2",
            time_model_id="tm1",
            type=ProcessTypeEnum.CapabilityProcesses,
            capability="C1",
        )
        process3 = CapabilityProcessData(
            ID="CP3",
            description="Capability Process 3",
            time_model_id="tm1",
            type=ProcessTypeEnum.CapabilityProcesses,
            capability="C2",  # different capability
        )
        
        hash1 = process1.hash(sample_production_system)
        hash2 = process2.hash(sample_production_system)
        hash3 = process3.hash(sample_production_system)
        
        # Same parameters should produce same hash
        assert hash1 == hash2
        # Different capability should produce different hash
        assert hash1 != hash3


class TestTransportProcessData:
    """Tests for TransportProcessData."""

    def test_creation_with_required_fields(self):
        """Test creating TransportProcessData with required fields."""
        process = TransportProcessData(
            ID="TP1",
            description="Transport Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.TransportProcesses,
        )
        assert process.ID == "TP1"
        assert process.time_model_id == "tm1"
        assert process.type == ProcessTypeEnum.TransportProcesses

    def test_with_loading_and_unloading_time_models(self):
        """Test creating TransportProcessData with loading/unloading time models."""
        process = TransportProcessData(
            ID="TP1",
            description="Transport Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.TransportProcesses,
            loading_time_model_id="tm_load",
            unloading_time_model_id="tm_unload",
        )
        assert process.loading_time_model_id == "tm_load"
        assert process.unloading_time_model_id == "tm_unload"


class TestReworkProcessData:
    """Tests for ReworkProcessData."""

    def test_creation_with_reworked_process_ids(self):
        """Test creating ReworkProcessData with reworked process IDs."""
        process = ReworkProcessData(
            ID="RP1",
            description="Rework Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.ReworkProcesses,
            reworked_process_ids=["P1", "P2"],
            blocking=True,
        )
        assert process.ID == "RP1"
        assert process.reworked_process_ids == ["P1", "P2"]
        assert process.blocking is True

    def test_non_blocking_rework(self):
        """Test creating non-blocking ReworkProcessData."""
        process = ReworkProcessData(
            ID="RP1",
            description="Rework Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.ReworkProcesses,
            reworked_process_ids=["P1"],
            blocking=False,
        )
        assert process.blocking is False


class TestCompoundProcessData:
    """Tests for CompoundProcessData."""

    def test_creation_with_process_ids(self):
        """Test creating CompoundProcessData with process IDs."""
        process = CompoundProcessData(
            ID="CP1",
            description="Compound Process 1",
            process_ids=["P1", "P2"],
            type=ProcessTypeEnum.CompoundProcesses,
        )
        assert process.ID == "CP1"
        assert process.process_ids == ["P1", "P2"]
        assert process.type == ProcessTypeEnum.CompoundProcesses


class TestRequiredCapabilityProcessData:
    """Tests for RequiredCapabilityProcessData."""

    def test_creation_with_capability(self):
        """Test creating RequiredCapabilityProcessData with capability."""
        process = RequiredCapabilityProcessData(
            ID="RCP1",
            description="Required Capability Process 1",
            type=ProcessTypeEnum.RequiredCapabilityProcesses,
            capability="C1",
        )
        assert process.ID == "RCP1"
        assert process.capability == "C1"
        assert process.type == ProcessTypeEnum.RequiredCapabilityProcesses


class TestLinkTransportProcessData:
    """Tests for LinkTransportProcessData."""

    def test_creation_with_links(self):
        """Test creating LinkTransportProcessData with links."""
        process = LinkTransportProcessData(
            ID="LTP1",
            description="Link Transport Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.LinkTransportProcesses,
            links=[["Resource1", "Node2"], ["Node2", "Resource1"]],
        )
        assert process.ID == "LTP1"
        assert process.links == [["Resource1", "Node2"], ["Node2", "Resource1"]]

    def test_default_capability(self):
        """Test that capability defaults to empty string."""
        process = LinkTransportProcessData(
            ID="LTP1",
            description="Link Transport Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.LinkTransportProcesses,
            links=[["Resource1", "Node2"]],
        )
        assert process.capability == ""

    def test_with_capability(self):
        """Test creating LinkTransportProcessData with capability."""
        process = LinkTransportProcessData(
            ID="LTP1",
            description="Link Transport Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.LinkTransportProcesses,
            links=[["Resource1", "Node2"]],
            capability="automated_transport",
        )
        assert process.capability == "automated_transport"


class TestProcessModelData:
    """Tests for ProcessModelData."""

    def test_creation_with_adjacency_matrix(self):
        """Test creating ProcessModelData with adjacency matrix."""
        process_model = ProcessModelData(
            ID="PM1",
            description="Process Model 1",
            type=ProcessTypeEnum.ProcessModels,
            adjacency_matrix={"P1": ["P2"], "P2": ["P3"], "P3": []},
        )
        assert process_model.ID == "PM1"
        assert process_model.adjacency_matrix == {"P1": ["P2"], "P2": ["P3"], "P3": []}

    def test_process_ids_property(self):
        """Test that process_ids property returns keys from adjacency_matrix."""
        process_model = ProcessModelData(
            ID="PM1",
            description="Process Model 1",
            type=ProcessTypeEnum.ProcessModels,
            adjacency_matrix={"P1": ["P2"], "P2": ["P3"], "P3": []},
        )
        assert set(process_model.process_ids) == {"P1", "P2", "P3"}

    def test_default_dependency_ids(self):
        """Test that dependency_ids defaults to empty list."""
        process_model = ProcessModelData(
            ID="PM1",
            description="Process Model 1",
            type=ProcessTypeEnum.ProcessModels,
            adjacency_matrix={"P1": []},
        )
        assert process_model.dependency_ids == []


class TestLoadingProcessData:
    """Tests for LoadingProcessData."""

    def test_creation_with_dependency_type(self):
        """Test creating LoadingProcessData with dependency type."""
        process = LoadingProcessData(
            ID="LP1",
            description="Loading Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.LoadingProcesses,
            dependency_type="before",
        )
        assert process.ID == "LP1"
        assert process.dependency_type == "before"

    def test_default_can_be_chained(self):
        """Test that can_be_chained defaults to True."""
        process = LoadingProcessData(
            ID="LP1",
            description="Loading Process 1",
            time_model_id="tm1",
            type=ProcessTypeEnum.LoadingProcesses,
            dependency_type="before",
        )
        assert process.can_be_chained is True

    def test_different_dependency_types(self):
        """Test creating LoadingProcessData with different dependency types."""
        for dep_type in ["before", "after", "parallel"]:
            process = LoadingProcessData(
                ID="LP1",
                description="Loading Process 1",
                time_model_id="tm1",
                type=ProcessTypeEnum.LoadingProcesses,
                dependency_type=dep_type,
            )
            assert process.dependency_type == dep_type




from __future__ import annotations

from typing import List, Optional
from uuid import uuid1

from pydantic import Field
from pydantic.dataclasses import dataclass

from prodsys.express import core
from prodsys.express.product import Product
from prodsys.models import order_data


@dataclass
class OrderedProduct(core.ExpressObject):
    """
    Class that represents an ordered product in an order.

    Args:
        product (Product): The product to order.
        quantity (int): Quantity of the product to order.
    """

    product: Product
    quantity: int

    def to_model(self) -> order_data.OrderedProductData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            order_data.OrderedProductData: An instance of the data object.
        """
        return order_data.OrderedProductData(
            product_type=self.product.ID,
            quantity=self.quantity,
        )


@dataclass
class Order(core.ExpressObject):
    """
    Class that represents an order.

    Args:
        ordered_products (List[OrderedProduct]): List of products to order.
        order_time (float): Time when the order was placed.
        due_time (Optional[float]): Time when the order should be due. If not specified, uses order_time.
        release_time (Optional[float]): Time when the order should be released. If not specified, uses order_time.
        priority (int): Priority of the order.
        ID (str): ID of the order.
    """

    ordered_products: List[OrderedProduct]
    order_time: float
    due_time: Optional[float] = None
    release_time: Optional[float] = None
    priority: int = 1
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    def to_model(self) -> order_data.OrderData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            order_data.OrderData: An instance of the data object.
        """
        return order_data.OrderData(
            ID=self.ID,
            ordered_products=[op.to_model() for op in self.ordered_products],
            order_time=self.order_time,
            due_time=self.due_time,
            release_time=self.release_time,
            priority=self.priority,
        )




from __future__ import annotations

from abc import ABC, abstractmethod

from functools import partial
from typing import TYPE_CHECKING, List, Optional, Tuple, Callable, Union

import numpy as np
from simpy import events

import gymnasium as gym
from gymnasium import spaces
from prodsys import adapters, runner
from prodsys.simulation import control, sim, observer, router

sim.VERBOSE = 0

if TYPE_CHECKING:
    from prodsys.simulation import resources, request


class AbstractRoutingControlEnv(gym.Env, ABC):
    """
    Abstract Gym environment for controlling a router of a production system with an reinforcement learning agent.
    This class defines the methods that need to be implemented in order to use a Reinforcement learning agent for production routing control.

    Args:
        adapter (adapters.ProductionSystemAdapter): The adapter.
        observation_space (Optional[spaces.Space], optional): The observation space of the environment.
        action_space (Optional[spaces.Space], optional): The action space of the environment.
        render_mode (Optional[str], optional): The render mode of the environment. Defaults to None.

    Attributes:
        adapter (adapters.ProductionSystemAdapter): The adapter.
        observation_space (Optional[spaces.Space], optional): The observation space of the environment.
        action_space (Optional[spaces.Space], optional): The action space of the environment.
        render_mode (Optional[str], optional): The render mode of the environment. Defaults to None.
        runner (runner.Runner): The runner of the adapter.
        router: (router.Router): The router of the adapter.
        possible_resources (List[resources.Resource]): The possible resources to route to.
        interrupt_simulation_event (events.Event): The event to interrupt the simulation when an agent interaction is needed.
        observers (List[observer.ResourceObserver]): The observers for the reosurces to route to.
        step_count (int): The number of steps taken in the environment.
        reward (float): The reward of the environment.
    """

    def __init__(
        self,
        adapter: adapters.ProductionSystemData,
        observation_space: Optional[spaces.Space] = None,
        action_space: Optional[spaces.Space] = None,
        render_mode: Optional[str] = None,
    ):
        self.adapter = adapter
        self.observation_space = observation_space
        self.action_space = action_space
        self.render_mode = render_mode

        self.runner = runner.Runner(production_system_data=self.adapter)

        self.router: router.Router = None
        self.possible_requests: List[request.Request] = []
        self.chosen_resource: Optional[resources.Resource] = None
        self.interrupt_simulation_event: events.Event = None
        self.observers: List[observer.ResourceObserver] = []
        self.step_count = 0
        self.reward = 0

    @abstractmethod
    def get_observation(self) -> np.ndarray:
        """
        Get observation of the environment.

        Returns:
            np.ndarray: The observation.
        """
        pass

    @abstractmethod
    def get_info(self) -> dict:
        """
        Get info of the environment.

        Returns:
            dict: The info.
        """
        pass

    @abstractmethod
    def get_reward(self, invalid_action: bool = False) -> float:
        """
        Get reward of the environment.

        Args:
            invalid_action (bool, optional): Whether the last action was invalid. Defaults to False.

        Returns:
            float: The reward.
        """
        pass

    @abstractmethod
    def get_termination_condition(self) -> bool:
        """
        Get termination condition of the environment.

        Returns:
            bool: The termination condition.
        """
        pass

    def reset(self, seed=None):
        """
        Reset env for new episode and run until first point of observation.

        Args:
            seed (Optional[int], optional): The seed for the environment. Defaults to None.
            options (Optional[dict], optional): The options for the environment. Defaults to None.
        """
        super().reset(seed=seed)

        self.runner.initialize_simulation()
        self.interrupt_simulation_event = events.Event(self.runner.env)
        self.chose_resource_event = events.Event(self.runner.env)
        # v1 runner creates and starts routers via RouterFactory.
        # For RL routing control, we keep the global router but override the routing heuristic
        # used when creating products so that routing decisions trigger agent interaction.
        self.router = self.runner.router_factory.global_system_router
        agent_routing_heuristic = partial(router.agent_routing_heuristic, self)
        self.runner.product_factory.routing_heuristic_override = agent_routing_heuristic
        # Also update already-created products (if any)
        for product in self.runner.product_factory.products.values():
            product.routing_heuristic = agent_routing_heuristic

        self.observers = []
        for resource in self.runner.resource_factory.all_resources.values():
            obs = observer.ResourceObserver(
                resource_factory=self.runner.resource_factory,
                product_factory=self.runner.product_factory,
                resource=resource,
            )
            self.observers.append(obs)

        self.runner.env.run_until(until=self.interrupt_simulation_event)
        self.interrupt_simulation_event = events.Event(self.runner.env)

        observation = self.get_observation()
        info = self.get_info()

        if self.render_mode == "human":
            self.render()

        return observation, info

    def set_possible_requests(self, requests: List[request.Request]):
        """
        Set possible requests for the RL agent environment.

        Args:
            resources (List[request.Request]): The possible requests to route.
        """
        self.possible_requests = requests

    def get_chosen_resource(self) -> resources.Resource:
        """
        Get the chosen resource of the RL agent for the router.

        Returns:
            resources.Resource: The chosen resource.
        """
        return self.chosen_resource

    def step(self, action: np.ndarray) -> Tuple[np.ndarray, float, bool, dict]:
        """
        Take a step in the environment.

        Args:
            action (np.ndarray): The output of the agent for actions.

        Returns:
            Tuple[np.ndarray, float, bool, dict]: The observation, reward, done, and info.
        """
        resource_index = np.argmax(action)

        self.chosen_resource = list(self.runner.resource_factory.all_resources.values())[
            resource_index
        ]
        if not self.chosen_resource.data.ID in [
            r.resource.data.ID for r in self.possible_requests
        ]:
            invalid_action = True
            self.chosen_resource = np.random.choice(self.possible_requests).resource
        else:
            invalid_action = False

        self.possible_requests.sort(
            key=lambda r: r.resource.data.ID == self.chosen_resource.data.ID,
            reverse=True,
        )

        self.runner.env.run_until(until=self.interrupt_simulation_event)
        self.step_count += 1
        self.interrupt_simulation_event = events.Event(self.runner.env)

        terminated = self.get_termination_condition()
        self.reward = self.get_reward(invalid_action)
        observation = self.get_observation()
        info = self.get_info()

        if self.render_mode == "human":
            self.render()

        return observation, self.reward, terminated, False, info

    def render(self):
        """
        Render the environment.
        """
        if self.render_mode == "human":
            pass

    def set_observation_space(self, observation_space: spaces.Space):
        """
        Set the observation space of the environment.

        Args:
            observation_space (spaces.Space): The observation space of the environment.
        """
        self.observation_space = observation_space

    def set_action_space(self, action_space: spaces.Space):
        """
        Set the action space of the environment.

        Args:
            action_space (spaces.Space): The action space of the environment.
        """
        self.action_space = action_space

    def set_binary_box_observation_space_from_shape(self, shape: Tuple[int, ...]):
        """
        Set the observation space of the environment to a binary box space.

        Args:
            shape (Tuple[int, ...]): The shape of the observation space.
        """
        self.observation_space = spaces.Box(0, 1, shape=shape, dtype=int)

    def set_binary_box_action_space_from_shape(self, shape: Tuple[int, ...]):
        """
        Set the action space of the environment to a binary box space.

        Args:
            shape (Tuple[int, ...]): The shape of the action space.
        """
        self.action_space = spaces.Box(0, 1, shape=shape, dtype=float)



"""
Capacity-based optimization algorithm.

This module provides a simple but effective optimization approach that generates
random configurations based on capacity analysis and selects the best performing ones.
It's particularly useful for fast optimization runs and initial solution generation.
"""

from __future__ import annotations
import logging
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field, ConfigDict
from prodsys.optimization.adapter_manipulation import (
    configuration_capacity_based_asserted,
)
from prodsys.optimization.optimization import evaluate
from prodsys.util.util import run_from_ipython

if run_from_ipython():
    from multiprocessing.pool import ThreadPool as Pool
else:
    from multiprocessing.pool import Pool

if TYPE_CHECKING:
    from prodsys.optimization.optimizer import Optimizer


class CapacityBasedHyperparameters(BaseModel):
    """
    Hyperparameters for capacity-based optimization.
    
    Args:
        num_solutions (int): Number of random solutions to generate and evaluate.
        cap_target (float): Target capacity utilization for resources (0.0 to 1.0).
                           Default is 0.65 (65% utilization).
        seed (int): Random seed for reproducibility. Default is 0.
        number_of_seeds (int): Number of simulation seeds to run per configuration.
                              Default is 1.
        number_of_processes (int): Number of processes to use for parallelization.
                                  Default is 1 (no parallelization).
    """
    num_solutions: int = Field(100, description="Number of solutions to generate and evaluate")
    cap_target: float = Field(0.65, description="Target capacity utilization (0.0-1.0)", ge=0.0, le=1.0)
    seed: int = Field(0, description="Random seed for reproducibility")
    number_of_seeds: int = Field(1, description="Number of simulation seeds per configuration")
    number_of_processes: int = Field(1, description="Number of processes for parallel execution")

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "num_solutions": 100,
                    "cap_target": 0.65,
                    "seed": 0,
                    "number_of_seeds": 1,
                    "number_of_processes": 4,
                },
            ]
        },
        extra="forbid",
    )


def _evaluate_configuration(args):
    """
    Wrapper function for parallel evaluation of configurations.
    
    Args:
        args: Tuple of (base_configuration, solution_dict, number_of_seeds, configuration, full_save)
    
    Returns:
        Tuple of (configuration, fitness_values, event_log_dict)
    """
    base_configuration, solution_dict, number_of_seeds, configuration, full_save = args
    fitness_values, event_log_dict = evaluate(
        base_scenario=base_configuration,
        solution_dict=solution_dict,
        number_of_seeds=number_of_seeds,
        adapter_object=configuration,
        full_save=full_save,
    )
    return configuration, fitness_values, event_log_dict


def capacity_based_optimization(optimizer: Optimizer) -> None:
    """
    Performs capacity-based optimization by generating random configurations
    based on capacity analysis and selecting the best performers.
    
    This is a simple but effective approach that:
    1. Generates N random configurations using capacity-based logic
    2. Evaluates each configuration via simulation (in parallel if specified)
    3. Returns the best configurations found
    
    The capacity-based generation ensures configurations are feasible and
    respect resource constraints while targeting specific utilization levels.
    
    Args:
        optimizer (Optimizer): The optimizer instance containing the base configuration
                              and hyperparameters.
    """
    hyperparameters: CapacityBasedHyperparameters = optimizer.hyperparameters
    base_configuration = optimizer.adapter
    
    logging.info(
        f"Starting capacity-based optimization with {hyperparameters.num_solutions} "
        f"solutions, cap_target={hyperparameters.cap_target}, "
        f"processes={hyperparameters.number_of_processes}"
    )
    
    # Set the generation identifier for tracking
    optimizer.optimization_cache_first_found_hashes.current_generation = "capacity_based"
    
    # PHASE 1: Generate all configurations
    logging.info("Generating configurations...")
    configurations = []
    generation_failures = 0
    
    for i in range(hyperparameters.num_solutions):
        try:
            configuration = configuration_capacity_based_asserted(
                baseline=base_configuration,
                cap_target=hyperparameters.cap_target
            )
            configurations.append(configuration)
        except Exception as e:
            generation_failures += 1
            logging.warning(
                f"Failed to generate configuration {i+1}/{hyperparameters.num_solutions}: {e}"
            )
            
            # If too many generation failures, stop early
            if generation_failures > hyperparameters.num_solutions * 0.5:
                logging.error(
                    f"More than 50% of configurations failed to generate ({generation_failures}/{i+1}). "
                    "Stopping optimization. Check configuration constraints."
                )
                break
    
    if not configurations:
        logging.error("No valid configurations were generated. Optimization aborted.")
        return
    
    logging.info(f"Generated {len(configurations)} valid configurations (failed: {generation_failures})")
    
    # PHASE 2: Evaluate all configurations (in parallel if specified)
    logging.info("Evaluating configurations...")
    
    # Prepare arguments for evaluation
    eval_args = [
        (
            base_configuration,
            optimizer.optimization_cache_first_found_hashes,
            hyperparameters.number_of_seeds,
            config,
            optimizer.full_save
        )
        for config in configurations
    ]
    
    # Execute evaluations (parallel or sequential)
    if hyperparameters.number_of_processes > 1:
        with Pool(hyperparameters.number_of_processes) as pool:
            results = pool.map(_evaluate_configuration, eval_args)
    else:
        results = [_evaluate_configuration(args) for args in eval_args]
    
    # PHASE 3: Save results
    solutions_evaluated = 0
    evaluation_failures = 0
    
    for configuration, fitness_values, event_log_dict in results:
        try:
            # Check if evaluation was successful (fitness_values would be None if cached or invalid)
            if fitness_values is None:
                # Configuration was either cached or invalid - still save it
                pass
            
            optimizer.save_optimization_step(
                fitness_values=fitness_values,
                configuration=configuration,
                event_log_dict=event_log_dict,
            )
            solutions_evaluated += 1
            
        except Exception as e:
            evaluation_failures += 1
            logging.warning(f"Failed to save evaluation result: {e}")
    
    logging.info(
        f"Capacity-based optimization completed. "
        f"Generated: {len(configurations)}, "
        f"Evaluated: {solutions_evaluated}, "
        f"Generation failures: {generation_failures}, "
        f"Evaluation failures: {evaluation_failures}"
    )




import pytest
import prodsys
from prodsys.models.production_system_data import ProductionSystemData
import prodsys.express as psx
from prodsys import runner


@pytest.fixture
def storage_simulation_adapter() -> ProductionSystemData:
    t1 = psx.FunctionTimeModel("exponential", 0.8, 0, "t1")

    p1 = psx.ProductionProcess(t1, "p1")
    p2 = psx.ProductionProcess(t1, "p2")

    t3 = psx.FunctionTimeModel("normal", 0.05, 0.01, ID="t3")

    tp = psx.TransportProcess(t3, "tp")

    machine = psx.Resource(
        [p1],
        [5, 0],
        3,
        ID="machine",
        internal_queue_size=20,
    )

    machine2 = psx.Resource(
        [p2],
        [9, 0],
        3,
        ID="machine2",
        internal_queue_size=20,
    )
    storage1 = psx.Store(ID="storage1", location=[6, 0], capacity=20)
    storage2 = psx.Store(ID="storage2", location=[11, 0], capacity=30)
    machine.buffers = [storage1]
    machine2.buffers = [storage2]

    transport = psx.Resource([tp], [0, 0], 1, ID="transport")

    product1 = psx.Product([p1, p2], tp, "product1")

    sink1 = psx.Sink(product1, [10, 0], "sink1")

    arrival_model_1 = psx.FunctionTimeModel("exponential", 1, ID="arrival_model_1")

    source1 = psx.Source(product1, arrival_model_1, [0, 0], ID="source_1")

    system = psx.ProductionSystem([machine, machine2, transport], [source1], [sink1])
    adapter = system.to_model()
    return adapter


def test_initialize_simulation(storage_simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=storage_simulation_adapter)
    runner_instance.initialize_simulation()


def test_hashing(storage_simulation_adapter: ProductionSystemData):
    hash_str = storage_simulation_adapter.hash()
    assert hash_str == "2470c17f770474e1dcc59f8f0fb25df6"


def test_run_simulation(storage_simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=storage_simulation_adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(2000)
    runner_instance.print_results()
    assert runner_instance.env.now == 2000
    post_processor = runner_instance.get_post_processor()
    for kpi in post_processor.throughput_and_output_KPIs:
        if kpi.name == "output":
            assert kpi.product_type == "product1"
            assert kpi.value > 1950 and kpi.value < 2060
    for kpi in post_processor.machine_state_KPIS:
        if kpi.name == "productive_time" and kpi.resource == "machine":
            assert kpi.value < 60 and kpi.value > 50

        if kpi.name == "productive_time" and kpi.resource == "transport":
            assert kpi.value > 35 and kpi.value < 50

    for kpi in post_processor.WIP_KPIs:
        if kpi.name == "WIP" and kpi.product_type == "product1":
            assert kpi.value > 2.5 and kpi.value < 3.5

    for kpi in post_processor.aggregated_throughput_time_KPIs:
        if kpi.name == "throughput_time":
            assert kpi.value > 2.0 and kpi.value < 3.0



"""
Tests for product_data module.
"""

import pytest
from prodsys.models.product_data import ProductData
from prodsys.models.processes_data import ProductionProcessData, ProcessTypeEnum
from prodsys.models.time_model_data import FunctionTimeModelData
from prodsys.models.source_data import RoutingHeuristic
from prodsys.models.production_system_data import ProductionSystemData
from prodsys.util.statistical_functions import FunctionTimeModelEnum


@pytest.fixture
def sample_time_model():
    """Create a sample time model for testing."""
    return FunctionTimeModelData(
        ID="tm1",
        description="Test time model",
        distribution_function=FunctionTimeModelEnum.Normal,
        location=20.0,
        scale=5.0,
    )


@pytest.fixture
def sample_transport_time_model():
    """Create a sample transport time model for testing."""
    return FunctionTimeModelData(
        ID="tm_transport",
        description="Transport time model",
        distribution_function=FunctionTimeModelEnum.Normal,
        location=1.0,
        scale=0.1,
    )


@pytest.fixture
def sample_process(sample_time_model):
    """Create a sample process for testing."""
    return ProductionProcessData(
        ID="P1",
        description="Process 1",
        time_model_id="tm1",
        type=ProcessTypeEnum.ProductionProcesses,
    )


@pytest.fixture
def sample_transport_process(sample_transport_time_model):
    """Create a sample transport process for testing."""
    from prodsys.models.processes_data import TransportProcessData
    
    return TransportProcessData(
        ID="TP1",
        description="Transport Process 1",
        time_model_id="tm_transport",
        type=ProcessTypeEnum.TransportProcesses,
    )


@pytest.fixture
def sample_production_system(sample_time_model, sample_transport_time_model, sample_process, sample_transport_process):
    """Create a minimal production system for testing."""
    return ProductionSystemData(
        time_model_data=[sample_time_model, sample_transport_time_model],
        process_data=[sample_process, sample_transport_process],
        resource_data=[],
        product_data=[],
        source_data=[],
        sink_data=[],
        state_data=[],
        port_data=[],
        primitive_data=[],
        dependency_data=[],
        node_data=[],
        scenario_data=None,
    )


class TestProductData:
    """Tests for ProductData."""

    def test_creation_with_sequential_processes(self):
        """Test creating ProductData with sequential processes (list)."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes=["P1", "P2", "P3"],
            transport_process="TP1",
        )
        assert product.ID == "Product_1"
        assert product.description == "Product 1"
        assert product.type == "Product_1"
        # Sequential list should be converted to adjacency matrix
        assert "P1" in product.processes
        assert product.processes["P1"] == ["P2"]
        assert product.processes["P2"] == ["P3"]
        assert product.processes["P3"] == []
        assert product.transport_process == "TP1"

    def test_creation_with_adjacency_matrix(self):
        """Test creating ProductData with adjacency matrix."""
        adjacency_matrix = {
            "P1": ["P2", "P3"],
            "P2": ["P3"],
            "P3": [],
        }
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes=adjacency_matrix,
            transport_process="TP1",
        )
        assert product.processes == adjacency_matrix

    def test_default_dependency_ids(self):
        """Test that dependency_ids defaults to empty list."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes={"P1": []},
            transport_process="TP1",
        )
        assert product.dependency_ids == []

    def test_with_dependency_ids(self):
        """Test creating ProductData with dependency IDs."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes={"P1": []},
            transport_process="TP1",
            dependency_ids=["dep1", "dep2"],
        )
        assert product.dependency_ids == ["dep1", "dep2"]

    def test_default_routing_heuristic(self):
        """Test that routing_heuristic defaults to None."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes={"P1": []},
            transport_process="TP1",
        )
        assert product.routing_heuristic is None

    def test_with_routing_heuristic(self):
        """Test creating ProductData with routing heuristic."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes={"P1": []},
            transport_process="TP1",
            routing_heuristic=RoutingHeuristic.shortest_queue,
        )
        assert product.routing_heuristic == RoutingHeuristic.shortest_queue

    def test_default_becomes_consumable(self):
        """Test that becomes_consumable defaults to False."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes={"P1": []},
            transport_process="TP1",
        )
        assert product.becomes_consumable is False

    def test_with_becomes_consumable(self):
        """Test creating ProductData with becomes_consumable=True."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes={"P1": []},
            transport_process="TP1",
            becomes_consumable=True,
        )
        assert product.becomes_consumable is True

    def test_type_synced_with_id(self):
        """Test that type is synced with ID if not provided."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            processes={"P1": []},
            transport_process="TP1",
        )
        assert product.type == "Product_1"

    def test_id_synced_with_type(self):
        """Test that ID is synced with type if type provided."""
        product = ProductData(
            description="Product 1",
            type="Product_1",
            processes={"P1": []},
            transport_process="TP1",
        )
        assert product.ID == "Product_1"

    def test_hash(self, sample_production_system, sample_process, sample_transport_process):
        """Test hash method."""
        product1 = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes={"P1": []},
            transport_process="TP1",
        )
        product2 = ProductData(
            ID="Product_2",
            description="Product 2",
            type="Product_2",
            processes={"P1": []},
            transport_process="TP1",
        )
        
        hash1 = product1.hash(sample_production_system)
        hash2 = product2.hash(sample_production_system)
        
        # Same processes and transport should produce same hash
        assert hash1 == hash2

    def test_inheritance_from_primitive_data(self):
        """Test that ProductData inherits from PrimitiveData."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes={"P1": []},
            transport_process="TP1",
        )
        from prodsys.models.primitives_data import PrimitiveData
        assert isinstance(product, PrimitiveData)
        assert hasattr(product, "type")
        assert hasattr(product, "transport_process")

    def test_sequential_to_adjacency_conversion(self):
        """Test that sequential list is correctly converted to adjacency matrix."""
        product = ProductData(
            ID="Product_1",
            description="Product 1",
            type="Product_1",
            processes=["P1", "P2", "P3"],
            transport_process="TP1",
        )
        assert product.processes["P1"] == ["P2"]
        assert product.processes["P2"] == ["P3"]
        assert product.processes["P3"] == []
        assert len(product.processes) == 3




from __future__ import annotations

from typing import Generator, TYPE_CHECKING

import logging

from prodsys.simulation import (
    sim,
    state,
    process,
)
from prodsys.models.dependency_data import DependencyType
from prodsys.simulation.entities.entity import EntityType

if TYPE_CHECKING:
    from prodsys.simulation import (
        process,
        state,
    )
    from prodsys.simulation import request as request_module
    from prodsys.simulation.entities import product


logger = logging.getLogger(__name__)


def get_process_time_for_lots(
        request: request_module.Request
) -> float:
    """
    Get the expected process time for a batch of requests.

    Args:
        request (request_module.Request): The request to get the process time for.

    Returns:
        float: The expected process time for the batch.
    """
    if not request.process:
        raise ValueError("Request has no process.")
    return request.process.time_model.get_next_time(
    )


class ProductionProcessHandler:
    """
    A production controller is responsible for controlling the processes of a production resource. The controller is requested by products requiring processes. The controller decides has a control policy that determines with which sequence requests are processed.
    """

    def __init__(self, env: sim.Environment) -> None:
        self.env = env
        self.resource = None

    def get_entities_of_request(
        self, process_request: request_module.Request
    ) -> Generator:
        """
        Get the next product for a process. The product is removed (get) from the input queues of the resource.

        Args:
            process_request (request_module.Request): The request to get the entities from.

        Returns:
            Generator: The generator yields when the product is taken from the queue (multiple events for multiple products, e.g. for a batch process or an assembly).
        """
        for entity in process_request.get_atomic_entities():
            entity.info.log_start_loading(process_request.resource, entity, self.env.now, process_request.origin_queue)
            yield from process_request.origin_queue.get(entity.data.ID)
            entity.info.log_end_loading(process_request.resource, entity, self.env.now, process_request.origin_queue)
        required_primitive_types = [dependency.data.required_entity for dependency in process_request.required_dependencies if dependency.data.dependency_type == DependencyType.TOOL or dependency.data.dependency_type == DependencyType.ASSEMBLY]
        for dependant_entity in process_request.entity.depended_entities:
            if dependant_entity.data.type not in required_primitive_types:
                continue
            dependant_entity.info.log_start_loading(process_request.resource, dependant_entity, self.env.now, process_request.origin_queue)
            yield from dependant_entity.current_locatable.get(dependant_entity.data.ID)
            dependant_entity.update_location(process_request.resource)
            dependant_entity.info.log_end_loading(process_request.resource, dependant_entity, self.env.now, dependant_entity.current_locatable)

    def put_entities_of_request(
        self, process_request: request_module.Request
    ) -> Generator:
        """
        Place a product to the output queue (put) of the resource.

        Args:
            process_request (request_module.Request): The request to place the product to.

        Returns:
            Generator: The generator yields when the product is placed in the queue (multiple events for multiple products, e.g. for a batch process or an assembly).
        """
        for entity in process_request.get_atomic_entities():
            entity.info.log_start_unloading(process_request.resource, entity, self.env.now, process_request.target_queue)
            yield from process_request.target_queue.put(entity.data)
            entity.info.log_end_unloading(process_request.resource, entity, self.env.now, process_request.target_queue)
        required_assembly_types = [dependency.data.required_entity for dependency in process_request.required_dependencies if dependency.data.dependency_type == DependencyType.ASSEMBLY]
        for dependant_entity in process_request.entity.depended_entities:
            if dependant_entity.data.type not in required_assembly_types:
                continue    
            dependant_entity.info.log_consumption(process_request.resource, dependant_entity, self.env.now)
        required_tool_types = [dependency.data.required_entity for dependency in process_request.required_dependencies if dependency.data.dependency_type == DependencyType.TOOL]
        for dependant_entity in process_request.entity.depended_entities:
            if dependant_entity.data.type not in required_tool_types:
                continue
            dependant_entity.current_locatable = process_request.entity._current_locatable
            dependant_entity.info.log_start_unloading(dependant_entity.current_locatable.resource, dependant_entity, self.env.now, dependant_entity.current_locatable)
            yield from dependant_entity.current_locatable.put(dependant_entity.data)
            dependant_entity.info.log_end_unloading(dependant_entity.current_locatable.resource, dependant_entity, self.env.now, dependant_entity.current_locatable)

    def handle_request(self, process_request: request_module.Request) -> Generator:
        """
        Start the next process with the following logic:

        1. Setup the resource for the process.
        2. Wait until the resource is free for the process.
        3. Retrieve the product from the queue.
        4. Run the process and wait until finished.
        5. Place the product in the output queue.

        Yields:
            Generator: The generator yields when the process is finished.
        """
        resource = process_request.get_resource()
        self.resource = resource
        process = process_request.get_process()

        # Take only dependencies of the main request of the lot
        if process_request.required_dependencies:
            yield process_request.request_dependencies()
        yield from resource.setup(process)
        resource_requests = []
        for _ in range(process_request.capacity_required):
            resource_request = resource.request()
            yield resource_request
            resource_requests.append(resource_request)
        

            
        # Now get all entities
        yield from self.get_entities_of_request(process_request)

        process_time = get_process_time_for_lots(process_request)
        resource.controller.mark_started_process(process_request.capacity_required)
        process_state_events = []
        for entity in process_request.get_atomic_entities():
            entity.update_location(process_request.resource)
            production_state: state.State = yield from resource.wait_for_free_process(
                process
            )
            production_state.reserved = True
            process_event = self.env.process(self.run_process(production_state,entity, process, process_time))
            process_state_events.append((process_event, production_state))
        for process_event, production_state in process_state_events:
            yield process_event
            # Store failure status on entity if available
            if hasattr(production_state.state_info, '_process_ok'):
                for entity in process_request.get_atomic_entities():
                    entity.last_process_failed = not production_state.state_info._process_ok
            production_state.process = None

        for resource_request in resource_requests:
            resource.release(resource_request)
        yield from self.put_entities_of_request(process_request)
        process_request.entity.update_location(process_request.target_queue)

        if process_request.entity.type == EntityType.LOT:
            process_request.entity.clear()

        buffer_placement_events = []
        for entity in process_request.get_atomic_entities():
            buffer_placement_event = entity.router.request_buffering(process_request)
            if buffer_placement_event:
                buffer_placement_events.append(buffer_placement_event)
        for buffer_placement_event in buffer_placement_events:
            yield buffer_placement_event

        self.resource.controller.mark_finished_process(process_request.capacity_required)
        process_request.entity.router.mark_finished_request(process_request)

    def run_process(
        self,
        input_state: state.State,
        target_product: product.Product,
        process: process.Process,
        process_time: float,
    ) -> Generator:
        """
        Run the process of a product. The process is started and the product is logged.

        Args:
            input_state (state.State): The production state of the process.
            target_product (product.Product): The product that is processed.
            process (process.Process): The process to run.
            process_time (float): The process time.
        """
        input_state.state_info.log_product(
            target_product, state.StateTypeEnum.production
        )
        input_state.process = self.env.process(input_state.process_state(time=process_time))  # type: ignore False
        input_state.reserved = False

        yield input_state.process



import pytest
import random
import numpy as np
import prodsys
from prodsys.optimization.adapter_manipulation import reset_transformations


@pytest.fixture(autouse=True, scope="function")
def cleanup_test_state():
    """Fixture to clean up global test state before and after each test to avoid test pollution.
    
    This ensures that:
    - Transformation operations added during tests are cleaned up
    - Logging state is reset to default (WARNING)
    - Random state is reset to a known default (seed=0) to ensure deterministic behavior
    - Route caches are cleared to prevent cross-test route pollution
    """
    # Reset before test to ensure clean state
    reset_transformations()
    prodsys.set_logging("WARNING")  # Reset to default logging level
    # Set a default seed to ensure deterministic behavior
    # Individual tests can override this by setting their own seed
    np.random.seed(0)
    random.seed(0)
    
    # Clear route caches that persist across tests
    # These are class-level/module-level caches that can cause test pollution
    try:
        from prodsys.simulation.route_finder import RouteFinder
        RouteFinder.clear_cache()
    except ImportError:
        pass
    
    yield
    
    # Reset after test to clean up for next test
    reset_transformations()
    prodsys.set_logging("WARNING")  # Reset to default logging level
    # Reset random state to default for next test
    np.random.seed(0)
    random.seed(0)
    
    # Clear route caches again after test
    try:
        from prodsys.simulation.route_finder import RouteFinder
        RouteFinder.clear_cache()
    except ImportError:
        pass

